import re
from enum import Enum
from typing import List, Literal, get_args


class ActorEventTypes(str, Enum):
    """Possible values of actor event type."""

    #: Info about resource usage of the actor
    SYSTEM_INFO = 'systemInfo'
    #: Sent when the actor is about to migrate
    MIGRATING = 'migrating'
    #: Sent when the actor should persist its state (every minute or when migrating)
    PERSIST_STATE = 'persistState'
    #: Sent when the actor is aborting
    ABORTING = 'aborting'


class ApifyEnvVars(str, Enum):
    """Possible Apify-specific environment variables."""

    # TODO: document these

    #: ACT_ID
    ACT_ID = 'APIFY_ACT_ID'
    #: ACT_RUN_ID
    ACT_RUN_ID = 'APIFY_ACT_RUN_ID'
    #: ACTOR_BUILD_ID
    ACTOR_BUILD_ID = 'APIFY_ACTOR_BUILD_ID'
    #: ACTOR_BUILD_NUMBER
    ACTOR_BUILD_NUMBER = 'APIFY_ACTOR_BUILD_NUMBER'
    #: ACTOR_EVENTS_WS_URL
    ACTOR_EVENTS_WS_URL = 'APIFY_ACTOR_EVENTS_WS_URL'
    #: ACTOR_ID
    ACTOR_ID = 'APIFY_ACTOR_ID'
    #: ACTOR_RUN_ID
    ACTOR_RUN_ID = 'APIFY_ACTOR_RUN_ID'
    #: ACTOR_TASK_ID
    ACTOR_TASK_ID = 'APIFY_ACTOR_TASK_ID'
    #: API_BASE_URL
    API_BASE_URL = 'APIFY_API_BASE_URL'
    #: API_PUBLIC_BASE_URL
    API_PUBLIC_BASE_URL = 'APIFY_API_PUBLIC_BASE_URL'
    #: CHROME_EXECUTABLE_PATH
    CHROME_EXECUTABLE_PATH = 'APIFY_CHROME_EXECUTABLE_PATH'
    #: CONTAINER_PORT
    CONTAINER_PORT = 'APIFY_CONTAINER_PORT'
    #: CONTAINER_URL
    CONTAINER_URL = 'APIFY_CONTAINER_URL'
    #: DEDICATED_CPUS
    DEDICATED_CPUS = 'APIFY_DEDICATED_CPUS'
    #: DEFAULT_BROWSER_PATH
    DEFAULT_BROWSER_PATH = 'APIFY_DEFAULT_BROWSER_PATH'
    #: DEFAULT_DATASET_ID
    DEFAULT_DATASET_ID = 'APIFY_DEFAULT_DATASET_ID'
    #: DEFAULT_KEY_VALUE_STORE_ID
    DEFAULT_KEY_VALUE_STORE_ID = 'APIFY_DEFAULT_KEY_VALUE_STORE_ID'
    #: DEFAULT_REQUEST_QUEUE_ID
    DEFAULT_REQUEST_QUEUE_ID = 'APIFY_DEFAULT_REQUEST_QUEUE_ID'
    #: DISABLE_BROWSER_SANDBOX
    DISABLE_BROWSER_SANDBOX = 'APIFY_DISABLE_BROWSER_SANDBOX'
    #: DISABLE_OUTDATED_WARNING
    DISABLE_OUTDATED_WARNING = 'APIFY_DISABLE_OUTDATED_WARNING'
    #: FACT
    FACT = 'APIFY_FACT'
    #: HEADLESS
    HEADLESS = 'APIFY_HEADLESS'
    #: INPUT_KEY
    INPUT_KEY = 'APIFY_INPUT_KEY'
    #: INPUT_SECRETS_PRIVATE_KEY_FILE
    INPUT_SECRETS_PRIVATE_KEY_FILE = 'APIFY_INPUT_SECRETS_PRIVATE_KEY_FILE'
    #: INPUT_SECRETS_PRIVATE_KEY_PASSPHRASE
    INPUT_SECRETS_PRIVATE_KEY_PASSPHRASE = 'APIFY_INPUT_SECRETS_PRIVATE_KEY_PASSPHRASE'
    #: IS_AT_HOME
    IS_AT_HOME = 'APIFY_IS_AT_HOME'
    #: LOCAL_STORAGE_DIR
    LOCAL_STORAGE_DIR = 'APIFY_LOCAL_STORAGE_DIR'
    #: LOG_FORMAT
    LOG_FORMAT = 'APIFY_LOG_FORMAT'
    #: LOG_LEVEL
    LOG_LEVEL = 'APIFY_LOG_LEVEL'
    #: MAX_USED_CPU_RATIO
    MAX_USED_CPU_RATIO = 'APIFY_MAX_USED_CPU_RATIO'
    #: MEMORY_MBYTES
    MEMORY_MBYTES = 'APIFY_MEMORY_MBYTES'
    #: META_ORIGIN
    META_ORIGIN = 'APIFY_META_ORIGIN'
    #: PERSIST_STORAGE
    PERSIST_STORAGE = 'APIFY_PERSIST_STORAGE'
    #: PROXY_HOSTNAME
    PROXY_HOSTNAME = 'APIFY_PROXY_HOSTNAME'
    #: PROXY_PASSWORD
    PROXY_PASSWORD = 'APIFY_PROXY_PASSWORD'
    #: PROXY_PORT
    PROXY_PORT = 'APIFY_PROXY_PORT'
    #: PROXY_STATUS_URL
    PROXY_STATUS_URL = 'APIFY_PROXY_STATUS_URL'
    #: SDK_LATEST_VERSION
    SDK_LATEST_VERSION = 'APIFY_SDK_LATEST_VERSION'
    #: STARTED_AT
    STARTED_AT = 'APIFY_STARTED_AT'
    #: TIMEOUT_AT
    TIMEOUT_AT = 'APIFY_TIMEOUT_AT'
    #: TOKEN
    TOKEN = 'APIFY_TOKEN'
    #: USER_ID
    USER_ID = 'APIFY_USER_ID'
    #: WORKFLOW_KEY
    WORKFLOW_KEY = 'APIFY_WORKFLOW_KEY'
    #: XVFB
    XVFB = 'APIFY_XVFB'

    # Extra ones not in @apify/consts:
    #: METAMORPH_AFTER_SLEEP_MILLIS
    METAMORPH_AFTER_SLEEP_MILLIS = 'APIFY_METAMORPH_AFTER_SLEEP_MILLIS'
    #: PERSIST_STATE_INTERVAL_MILLIS
    PERSIST_STATE_INTERVAL_MILLIS = 'APIFY_PERSIST_STATE_INTERVAL_MILLIS'
    #: PURGE_ON_START
    PURGE_ON_START = 'APIFY_PURGE_ON_START'
    #: SYSTEM_INFO_INTERVAL_MILLIS
    SYSTEM_INFO_INTERVAL_MILLIS = 'APIFY_SYSTEM_INFO_INTERVAL_MILLIS'


_INTEGER_ENV_VARS_TYPE = Literal[
    ApifyEnvVars.CONTAINER_PORT,
    ApifyEnvVars.DEDICATED_CPUS,
    ApifyEnvVars.LOG_LEVEL,
    ApifyEnvVars.MEMORY_MBYTES,
    ApifyEnvVars.METAMORPH_AFTER_SLEEP_MILLIS,
    ApifyEnvVars.PERSIST_STATE_INTERVAL_MILLIS,
    ApifyEnvVars.PROXY_PORT,
    ApifyEnvVars.SYSTEM_INFO_INTERVAL_MILLIS,
]

INTEGER_ENV_VARS: List[_INTEGER_ENV_VARS_TYPE] = list(get_args(_INTEGER_ENV_VARS_TYPE))

_FLOAT_ENV_VARS_TYPE = Literal[
    ApifyEnvVars.MAX_USED_CPU_RATIO,
]

FLOAT_ENV_VARS: List[_FLOAT_ENV_VARS_TYPE] = list(get_args(_FLOAT_ENV_VARS_TYPE))

_BOOL_ENV_VARS_TYPE = Literal[
    ApifyEnvVars.DISABLE_BROWSER_SANDBOX,
    ApifyEnvVars.DISABLE_OUTDATED_WARNING,
    ApifyEnvVars.HEADLESS,
    ApifyEnvVars.IS_AT_HOME,
    ApifyEnvVars.PERSIST_STORAGE,
    ApifyEnvVars.PURGE_ON_START,
    ApifyEnvVars.XVFB,
]

BOOL_ENV_VARS: List[_BOOL_ENV_VARS_TYPE] = list(get_args(_BOOL_ENV_VARS_TYPE))

_DATETIME_ENV_VARS_TYPE = Literal[
    ApifyEnvVars.STARTED_AT,
    ApifyEnvVars.TIMEOUT_AT,
]

DATETIME_ENV_VARS: List[_DATETIME_ENV_VARS_TYPE] = list(get_args(_DATETIME_ENV_VARS_TYPE))

_STRING_ENV_VARS_TYPE = Literal[
    ApifyEnvVars.ACT_ID,
    ApifyEnvVars.ACT_RUN_ID,
    ApifyEnvVars.ACTOR_BUILD_ID,
    ApifyEnvVars.ACTOR_BUILD_NUMBER,
    ApifyEnvVars.ACTOR_EVENTS_WS_URL,
    ApifyEnvVars.ACTOR_ID,
    ApifyEnvVars.ACTOR_RUN_ID,
    ApifyEnvVars.ACTOR_TASK_ID,
    ApifyEnvVars.API_BASE_URL,
    ApifyEnvVars.API_PUBLIC_BASE_URL,
    ApifyEnvVars.CHROME_EXECUTABLE_PATH,
    ApifyEnvVars.CONTAINER_URL,
    ApifyEnvVars.DEFAULT_BROWSER_PATH,
    ApifyEnvVars.DEFAULT_DATASET_ID,
    ApifyEnvVars.DEFAULT_KEY_VALUE_STORE_ID,
    ApifyEnvVars.DEFAULT_REQUEST_QUEUE_ID,
    ApifyEnvVars.FACT,
    ApifyEnvVars.INPUT_KEY,
    ApifyEnvVars.INPUT_SECRETS_PRIVATE_KEY_FILE,
    ApifyEnvVars.INPUT_SECRETS_PRIVATE_KEY_PASSPHRASE,
    ApifyEnvVars.LOCAL_STORAGE_DIR,
    ApifyEnvVars.LOG_FORMAT,
    ApifyEnvVars.META_ORIGIN,
    ApifyEnvVars.PROXY_HOSTNAME,
    ApifyEnvVars.PROXY_PASSWORD,
    ApifyEnvVars.PROXY_STATUS_URL,
    ApifyEnvVars.SDK_LATEST_VERSION,
    ApifyEnvVars.TOKEN,
    ApifyEnvVars.USER_ID,
    ApifyEnvVars.WORKFLOW_KEY,
]

STRING_ENV_VARS: List[_STRING_ENV_VARS_TYPE] = list(get_args(_STRING_ENV_VARS_TYPE))


class ActorExitCodes(int, Enum):
    """Usual actor exit codes."""

    #: The actor finished successfully
    SUCCESS = 0

    #: The main function of the actor threw an Exception
    ERROR_USER_FUNCTION_THREW = 91


class _StorageTypes(str, Enum):
    """Possible Apify storage types."""

    DATASET = 'Dataset'
    KEY_VALUE_STORE = 'Key-value store'
    REQUEST_QUEUE = 'Request queue'


DEFAULT_API_PARAM_LIMIT = 1000

REQUEST_ID_LENGTH = 15

REQUEST_QUEUE_HEAD_MAX_LIMIT = 1000

EVENT_LISTENERS_TIMEOUT_SECS = 5

BASE64_REGEXP = '[-A-Za-z0-9+/]*={0,3}'
ENCRYPTED_INPUT_VALUE_PREFIX = 'ENCRYPTED_VALUE'
ENCRYPTED_INPUT_VALUE_REGEXP = re.compile(f'^{ENCRYPTED_INPUT_VALUE_PREFIX}:({BASE64_REGEXP}):({BASE64_REGEXP})$')

MAX_PAYLOAD_SIZE_BYTES = 9437184  # 9MB



import json
import logging
import textwrap
import traceback
from typing import Any, Dict

from colorama import Fore, Style, just_fix_windows_console

from ._utils import ignore_docs

just_fix_windows_console()


# Name of the logger used throughout the library (resolves to 'apify')
logger_name = __name__.split('.')[0]

# Logger used throughout the library
logger = logging.getLogger(logger_name)


_LOG_LEVEL_COLOR = {
    logging.DEBUG: Fore.BLUE,
    logging.INFO: Fore.GREEN,
    logging.WARNING: Fore.YELLOW,
    logging.ERROR: Fore.RED,
    logging.CRITICAL: Fore.RED,
}

_LOG_LEVEL_SHORT_ALIAS = {
    logging.DEBUG: 'DEBUG',
    logging.INFO: 'INFO ',
    logging.WARNING: 'WARN ',
    logging.ERROR: 'ERROR',
}

# So that all the log messages have the same alignment
_LOG_MESSAGE_INDENT = ' ' * 6


class ActorLogFormatter(logging.Formatter):
    """Log formatter that prints out the log message nicely formatted, with colored level and stringified extra fields.

    It formats the log records so that they:
    - start with the level (colorized, and padded to 5 chars so that it is nicely aligned)
    - then have the actual log message, if it's multiline then it's nicely indented
    - then have the stringified extra log fields
    - then, if an exception is a part of the log record, prints the formatted exception.
    """

    # The fields that are added to the log record with `logger.log(..., extra={...})`
    # are just merged in the log record with the other log record properties, and you can't get them in some nice, isolated way.
    # So, to get the extra fields, we just compare all the properties present in the log record
    # with properties present in an empty log record,
    # and extract all the extra ones not present in the empty log record
    empty_record = logging.LogRecord('dummy', 0, 'dummy', 0, 'dummy', None, None)

    def _get_extra_fields(self, record: logging.LogRecord) -> Dict[str, Any]:
        extra_fields: Dict[str, Any] = {}
        for key, value in record.__dict__.items():
            if key not in self.empty_record.__dict__:
                extra_fields[key] = value

        return extra_fields

    @ignore_docs
    def format(self, record: logging.LogRecord) -> str:
        """Format the log record nicely.

        This formats the log record so that it:
        - starts with the level (colorized, and padded to 5 chars so that it is nicely aligned)
        - then has the actual log message, if it's multiline then it's nicely indented
        - then has the stringified extra log fields
        - then, if an exception is a part of the log record, prints the formatted exception.
        """
        # Colorize the log level, and shorten it to 6 chars tops
        level_color_code = _LOG_LEVEL_COLOR.get(record.levelno, '')
        level_short_alias = _LOG_LEVEL_SHORT_ALIAS.get(record.levelno, record.levelname)
        level_string = f'{level_color_code}{level_short_alias}{Style.RESET_ALL} '

        # Format the exception, if there is some
        # Basically just print the traceback and indent it a bit
        exception_string = ''
        if record.exc_info:
            exc_info = record.exc_info
            record.exc_info = None
            exception_string = ''.join(traceback.format_exception(*exc_info)).rstrip()
            exception_string = '\n' + textwrap.indent(exception_string, _LOG_MESSAGE_INDENT)

        # Format the extra log record fields, if there were some
        # Just stringify them to JSON and color them gray
        extra_string = ''
        extra = self._get_extra_fields(record)
        if extra:
            extra_string = f' {Fore.LIGHTBLACK_EX}({json.dumps(extra, ensure_ascii=False, default=str)}){Style.RESET_ALL}'

        # Format the actual log message, and indent everything but the first line
        log_string = super().format(record)
        log_string = textwrap.indent(log_string, _LOG_MESSAGE_INDENT).lstrip()

        return f'{level_string}{log_string}{extra_string}{exception_string}'



from typing import Optional

from ._utils import _fetch_and_parse_env_var
from .consts import ApifyEnvVars


class Configuration:
    """A class for specifying the configuration of an actor.

    Can be used either globally via `Configuration.get_global_configuration()`,
    or it can be specific to each `Actor` instance on the `actor.config` property.
    """

    _default_instance: Optional['Configuration'] = None

    def __init__(
        self,
        *,
        api_base_url: Optional[str] = None,
        api_public_base_url: Optional[str] = None,
        container_port: Optional[int] = None,
        container_url: Optional[str] = None,
        default_dataset_id: Optional[str] = None,
        default_key_value_store_id: Optional[str] = None,
        default_request_queue_id: Optional[str] = None,
        input_key: Optional[str] = None,
        max_used_cpu_ratio: Optional[float] = None,
        metamorph_after_sleep_millis: Optional[int] = None,
        persist_state_interval_millis: Optional[int] = None,
        persist_storage: Optional[bool] = None,
        proxy_hostname: Optional[str] = None,
        proxy_password: Optional[str] = None,
        proxy_port: Optional[int] = None,
        proxy_status_url: Optional[str] = None,
        purge_on_start: Optional[bool] = None,
        token: Optional[str] = None,
        system_info_interval_millis: Optional[int] = None,
    ) -> None:
        """Create a `Configuration` instance.

        All the parameters are loaded by default from environment variables when running on the Apify platform.
        You can override them here in the Configuration constructor, which might be useful for local testing of your actors.

        Args:
            api_base_url (str, optional): The URL of the Apify API.
                This is the URL actually used for connecting to the API, so it can contain an IP address when running in a container on the platform.
            api_public_base_url (str, optional): The public URL of the Apify API.
                This will always contain the public URL of the API, even when running in a container on the platform.
                Useful for generating shareable URLs to key-value store records or datasets.
            container_port (int, optional): The port on which the container can listen for HTTP requests.
            container_url (str, optional): The URL on which the container can listen for HTTP requests.
            default_dataset_id (str, optional): The ID of the default dataset for the actor.
            default_key_value_store_id (str, optional): The ID of the default key-value store for the actor.
            default_request_queue_id (str, optional): The ID of the default request queue for the actor.
            input_key (str, optional): The key of the input record in the actor's default key-value store
            max_used_cpu_ratio (float, optional): The CPU usage above which the SYSTEM_INFO event will report the CPU is overloaded.
            metamorph_after_sleep_millis (int, optional): How long should the actor sleep after calling metamorph.
            persist_state_interval_millis (int, optional): How often should the actor emit the PERSIST_STATE event.
            persist_storage (bool, optional): Whether the actor should persist its used storages to the filesystem when running locally.
            proxy_hostname (str, optional): The hostname of Apify Proxy.
            proxy_password (str, optional): The password for Apify Proxy.
            proxy_port (str, optional): The port of Apify Proxy.
            proxy_status_url (str, optional): The URL on which the Apify Proxy status page is available.
            purge_on_start (str, optional): Whether the actor should purge its default storages on startup, when running locally.
            token (str, optional): The API token for the Apify API this actor should use.
            system_info_interval_millis (str, optional): How often should the actor emit the SYSTEM_INFO event when running locally.
        """
        # TODO: Document all these members
        self.actor_build_id = _fetch_and_parse_env_var(ApifyEnvVars.ACTOR_BUILD_ID)
        self.actor_build_number = _fetch_and_parse_env_var(ApifyEnvVars.ACTOR_BUILD_NUMBER)
        self.actor_events_ws_url = _fetch_and_parse_env_var(ApifyEnvVars.ACTOR_EVENTS_WS_URL)
        self.actor_id = _fetch_and_parse_env_var(ApifyEnvVars.ACTOR_ID)
        self.actor_run_id = _fetch_and_parse_env_var(ApifyEnvVars.ACTOR_RUN_ID)
        self.actor_task_id = _fetch_and_parse_env_var(ApifyEnvVars.ACTOR_TASK_ID)
        self.api_base_url = api_base_url or _fetch_and_parse_env_var(ApifyEnvVars.API_BASE_URL, 'https://api.apify.com')
        self.api_public_base_url = api_public_base_url or _fetch_and_parse_env_var(ApifyEnvVars.API_PUBLIC_BASE_URL, 'https://api.apify.com')
        self.chrome_executable_path = _fetch_and_parse_env_var(ApifyEnvVars.CHROME_EXECUTABLE_PATH)
        self.container_port = container_port or _fetch_and_parse_env_var(ApifyEnvVars.CONTAINER_PORT, 4321)
        self.container_url = container_url or _fetch_and_parse_env_var(ApifyEnvVars.CONTAINER_URL, 'http://localhost:4321')
        self.dedicated_cpus = _fetch_and_parse_env_var(ApifyEnvVars.DEDICATED_CPUS)
        self.default_browser_path = _fetch_and_parse_env_var(ApifyEnvVars.DEFAULT_BROWSER_PATH)
        self.default_dataset_id = default_dataset_id or _fetch_and_parse_env_var(ApifyEnvVars.DEFAULT_DATASET_ID, 'default')
        self.default_key_value_store_id = default_key_value_store_id or _fetch_and_parse_env_var(ApifyEnvVars.DEFAULT_KEY_VALUE_STORE_ID, 'default')
        self.default_request_queue_id = default_request_queue_id or _fetch_and_parse_env_var(ApifyEnvVars.DEFAULT_REQUEST_QUEUE_ID, 'default')
        self.disable_browser_sandbox = _fetch_and_parse_env_var(ApifyEnvVars.DISABLE_BROWSER_SANDBOX, False)
        self.headless = _fetch_and_parse_env_var(ApifyEnvVars.HEADLESS, True)
        self.input_key = input_key or _fetch_and_parse_env_var(ApifyEnvVars.INPUT_KEY, 'INPUT')
        self.input_secrets_private_key_file = _fetch_and_parse_env_var(ApifyEnvVars.INPUT_SECRETS_PRIVATE_KEY_FILE)
        self.input_secrets_private_key_passphrase = _fetch_and_parse_env_var(ApifyEnvVars.INPUT_SECRETS_PRIVATE_KEY_PASSPHRASE)
        self.is_at_home = _fetch_and_parse_env_var(ApifyEnvVars.IS_AT_HOME, False)
        self.max_used_cpu_ratio = max_used_cpu_ratio or _fetch_and_parse_env_var(ApifyEnvVars.MAX_USED_CPU_RATIO, 0.95)
        self.memory_mbytes = _fetch_and_parse_env_var(ApifyEnvVars.MEMORY_MBYTES)
        self.meta_origin = _fetch_and_parse_env_var(ApifyEnvVars.META_ORIGIN)
        self.metamorph_after_sleep_millis = metamorph_after_sleep_millis or _fetch_and_parse_env_var(ApifyEnvVars.METAMORPH_AFTER_SLEEP_MILLIS, 300000)  # noqa: E501
        self.persist_state_interval_millis = persist_state_interval_millis or _fetch_and_parse_env_var(ApifyEnvVars.PERSIST_STATE_INTERVAL_MILLIS, 60000)  # noqa: E501
        self.persist_storage = persist_storage or _fetch_and_parse_env_var(ApifyEnvVars.PERSIST_STORAGE, True)
        self.proxy_hostname = proxy_hostname or _fetch_and_parse_env_var(ApifyEnvVars.PROXY_HOSTNAME, 'proxy.apify.com')
        self.proxy_password = proxy_password or _fetch_and_parse_env_var(ApifyEnvVars.PROXY_PASSWORD)
        self.proxy_port = proxy_port or _fetch_and_parse_env_var(ApifyEnvVars.PROXY_PORT, 8000)
        self.proxy_status_url = proxy_status_url or _fetch_and_parse_env_var(ApifyEnvVars.PROXY_STATUS_URL, 'http://proxy.apify.com')
        self.purge_on_start = purge_on_start or _fetch_and_parse_env_var(ApifyEnvVars.PURGE_ON_START, False)
        self.started_at = _fetch_and_parse_env_var(ApifyEnvVars.STARTED_AT)
        self.timeout_at = _fetch_and_parse_env_var(ApifyEnvVars.TIMEOUT_AT)
        self.token = token or _fetch_and_parse_env_var(ApifyEnvVars.TOKEN)
        self.user_id = _fetch_and_parse_env_var(ApifyEnvVars.USER_ID)
        self.xvfb = _fetch_and_parse_env_var(ApifyEnvVars.XVFB, False)
        self.system_info_interval_millis = system_info_interval_millis or _fetch_and_parse_env_var(ApifyEnvVars.SYSTEM_INFO_INTERVAL_MILLIS, 60000)

    @classmethod
    def _get_default_instance(cls) -> 'Configuration':
        if cls._default_instance is None:
            cls._default_instance = cls()

        return cls._default_instance

    @classmethod
    def get_global_configuration(cls) -> 'Configuration':
        """Retrive the global configuration.

        The global configuration applies when you call actor methods via their static versions, e.g. `Actor.init()`.
        Also accessible via `Actor.config`.
        """
        return cls._get_default_instance()



from typing import Any, AsyncIterator, NamedTuple, Optional, TypedDict, TypeVar, Union, overload

from apify_client import ApifyClientAsync
from apify_client.clients import KeyValueStoreClientAsync, KeyValueStoreCollectionClientAsync

from .._memory_storage import MemoryStorageClient
from .._memory_storage.resource_clients import KeyValueStoreClient, KeyValueStoreCollectionClient
from .._utils import _wrap_internal, ignore_docs
from ..config import Configuration
from .base_storage import BaseStorage

T = TypeVar('T')
IterateKeysInfo = TypedDict('IterateKeysInfo', {'size': int})
IterateKeysTuple = NamedTuple('IterateKeysTuple', [('key', str), ('info', IterateKeysInfo)])


class KeyValueStore(BaseStorage):
    """The `KeyValueStore` class represents a key-value store.

    You can imagine it as a simple data storage that is used
    for saving and reading data records or files. Each data record is
    represented by a unique key and associated with a MIME content type.

    Do not instantiate this class directly, use the `Actor.open_key_value_store()` function instead.

    Each crawler run is associated with a default key-value store, which is created exclusively
    for the run. By convention, the crawler input and output are stored into the
    default key-value store under the `INPUT` and `OUTPUT` key, respectively.
    Typically, input and output are JSON files, although it can be any other format.
    To access the default key-value store directly, you can use the
    `KeyValueStore.get_value` and `KeyValueStore.set_value` convenience functions.

    `KeyValueStore` stores its data either on local disk or in the Apify cloud,
    depending on whether the `APIFY_LOCAL_STORAGE_DIR` or `APIFY_TOKEN` environment variables are set.

    If the `APIFY_LOCAL_STORAGE_DIR` environment variable is set, the data is stored in
    the local directory in the following files:
    ```
    {APIFY_LOCAL_STORAGE_DIR}/key_value_stores/{STORE_ID}/{INDEX}.{EXT}
    ```
    Note that `{STORE_ID}` is the name or ID of the key-value store. The default key-value store has ID: `default`,
    unless you override it by setting the `APIFY_DEFAULT_KEY_VALUE_STORE_ID` environment variable.
    The `{KEY}` is the key of the record and `{EXT}` corresponds to the MIME content type of the data value.

    If the `APIFY_TOKEN` environment variable is set but `APIFY_LOCAL_STORAGE_DIR` is not, the data is stored in the
    [Apify Key-value store](https://docs.apify.com/storage/key-value-store) cloud storage.
    """

    _id: str
    _name: Optional[str]
    _key_value_store_client: Union[KeyValueStoreClientAsync, KeyValueStoreClient]

    @ignore_docs
    def __init__(self, id: str, name: Optional[str], client: Union[ApifyClientAsync, MemoryStorageClient], config: Configuration) -> None:
        """Create a `KeyValueStore` instance.

        Do not use the constructor directly, use the `Actor.open_key_value_store()` function instead.

        Args:
            id (str): ID of the key-value store.
            name (str, optional): Name of the key-value store.
            client (ApifyClientAsync or MemoryStorageClient): The storage client which should be used.
            config (Configuration): The configuration which should be used.
        """
        super().__init__(id=id, name=name, client=client, config=config)

        self.get_value = _wrap_internal(self._get_value_internal, self.get_value)  # type: ignore
        self.set_value = _wrap_internal(self._set_value_internal, self.set_value)  # type: ignore
        self.get_public_url = _wrap_internal(self._get_public_url_internal, self.get_public_url)  # type: ignore
        self._id = id
        self._name = name
        self._key_value_store_client = client.key_value_store(self._id)

    @classmethod
    def _get_human_friendly_label(cls) -> str:
        return 'Key-value store'

    @classmethod
    def _get_default_id(cls, config: Configuration) -> str:
        return config.default_key_value_store_id

    @classmethod
    def _get_single_storage_client(
        cls,
        id: str,
        client: Union[ApifyClientAsync, MemoryStorageClient],
    ) -> Union[KeyValueStoreClientAsync, KeyValueStoreClient]:
        return client.key_value_store(id)

    @classmethod
    def _get_storage_collection_client(
        cls,
        client: Union[ApifyClientAsync, MemoryStorageClient],
    ) -> Union[KeyValueStoreCollectionClientAsync, KeyValueStoreCollectionClient]:
        return client.key_value_stores()

    @overload
    @classmethod
    async def get_value(cls, key: str) -> Any:
        ...

    @overload
    @classmethod
    async def get_value(cls, key: str, default_value: T) -> T:
        ...

    @overload
    @classmethod
    async def get_value(cls, key: str, default_value: Optional[T] = None) -> Optional[T]:
        ...

    @classmethod
    async def get_value(cls, key: str, default_value: Optional[T] = None) -> Optional[T]:
        """Get a value from the key-value store.

        Args:
            key (str): Key of the record to retrieve.
            default_value (Any, optional): Default value returned in case the record does not exist.

        Returns:
            Any: The value associated with the given key. `default_value` is used in case the record does not exist.
        """
        store = await cls.open()
        return await store.get_value(key, default_value)

    async def _get_value_internal(self, key: str, default_value: Optional[T] = None) -> Optional[T]:
        record = await self._key_value_store_client.get_record(key)
        return record['value'] if record else default_value

    async def iterate_keys(self, exclusive_start_key: Optional[str] = None) -> AsyncIterator[IterateKeysTuple]:
        """Iterate over the keys in the key-value store.

        Args:
            exclusive_start_key (str, optional): All keys up to this one (including) are skipped from the result.

        Yields:
            IterateKeysTuple: A tuple `(key, info)`,
                where `key` is the record key, and `info` is an object that contains a single property `size`
                indicating size of the record in bytes.
        """
        while True:
            list_keys = await self._key_value_store_client.list_keys(exclusive_start_key=exclusive_start_key)
            for item in list_keys['items']:
                yield IterateKeysTuple(item['key'], {'size': item['size']})

            if not list_keys['isTruncated']:
                break
            exclusive_start_key = list_keys['nextExclusiveStartKey']

    @classmethod
    async def set_value(cls, key: str, value: Optional[T], content_type: Optional[str] = None) -> None:
        """Set or delete a value in the key-value store.

        Args:
            key (str): The key under which the value should be saved.
            value (Any, optional): The value to save. If the value is `None`, the corresponding key-value pair will be deleted.
            content_type (str, optional): The content type of the saved value.
        """
        store = await cls.open()
        return await store.set_value(key, value, content_type)

    async def _set_value_internal(self, key: str, value: Optional[T], content_type: Optional[str] = None) -> None:
        if value is None:
            return await self._key_value_store_client.delete_record(key)

        return await self._key_value_store_client.set_record(key, value, content_type)

    @classmethod
    async def get_public_url(cls, key: str) -> str:
        """Get a URL for the given key that may be used to publicly access the value in the remote key-value store.

        Args:
            key (str): The key for which the URL should be generated.
        """
        store = await cls.open()
        return await store.get_public_url(key)

    async def _get_public_url_internal(self, key: str) -> str:
        if not isinstance(self._key_value_store_client, KeyValueStoreClientAsync):
            raise RuntimeError('Cannot generate a public URL for this key-value store as it is not on the Apify Platform!')

        public_api_url = self._config.api_public_base_url

        return f'{public_api_url}/v2/key-value-stores/{self._id}/records/{key}'

    async def drop(self) -> None:
        """Remove the key-value store either from the Apify cloud storage or from the local directory."""
        await self._key_value_store_client.delete()
        self._remove_from_cache()

    @classmethod
    async def open(
        cls,
        *,
        id: Optional[str] = None,
        name: Optional[str] = None,
        force_cloud: bool = False,
        config: Optional[Configuration] = None,
    ) -> 'KeyValueStore':
        """Open a key-value store.

        Key-value stores are used to store records or files, along with their MIME content type.
        The records are stored and retrieved using a unique key.
        The actual data is stored either on a local filesystem or in the Apify cloud.

        Args:
            id (str, optional): ID of the key-value store to be opened.
                If neither `id` nor `name` are provided, the method returns the default key-value store associated with the actor run.
                If the key-value store with the given ID does not exist, it raises an error.
            name (str, optional): Name of the key-value store to be opened.
                If neither `id` nor `name` are provided, the method returns the default key-value store associated with the actor run.
                If the key-value store with the given name does not exist, it is created.
            force_cloud (bool, optional): If set to True, it will open a key-value store on the Apify Platform even when running the actor locally.
                Defaults to False.
            config (Configuration, optional): A `Configuration` instance, uses global configuration if omitted.

        Returns:
            KeyValueStore: An instance of the `KeyValueStore` class for the given ID or name.
        """
        return await super().open(id=id, name=name, force_cloud=force_cloud, config=config)



import asyncio
from collections import OrderedDict
from datetime import datetime, timezone
from typing import Coroutine, Dict, Optional
from typing import OrderedDict as OrderedDictType
from typing import Set, Union

from apify_client import ApifyClientAsync
from apify_client.clients import RequestQueueClientAsync, RequestQueueCollectionClientAsync

from .._crypto import _crypto_random_object_id
from .._memory_storage import MemoryStorageClient
from .._memory_storage.resource_clients import RequestQueueClient, RequestQueueCollectionClient
from .._utils import LRUCache, _budget_ow, _unique_key_to_request_id, ignore_docs
from ..config import Configuration
from ..consts import REQUEST_QUEUE_HEAD_MAX_LIMIT
from ..log import logger
from .base_storage import BaseStorage

MAX_CACHED_REQUESTS = 1_000_000

# When requesting queue head we always fetch requestsInProgressCount * QUERY_HEAD_BUFFER number of requests.
QUERY_HEAD_MIN_LENGTH = 100

QUERY_HEAD_BUFFER = 3

# If queue was modified (request added/updated/deleted) before more than API_PROCESSED_REQUESTS_DELAY_MILLIS
# then we assume the get head operation to be consistent.
API_PROCESSED_REQUESTS_DELAY_MILLIS = 10_000

# How many times we try to get queue head with queueModifiedAt older than API_PROCESSED_REQUESTS_DELAY_MILLIS.
MAX_QUERIES_FOR_CONSISTENCY = 6

# This number must be large enough so that processing of all these requests cannot be done in
# a time lower than expected maximum latency of DynamoDB, but low enough not to waste too much memory.
RECENTLY_HANDLED_CACHE_SIZE = 1000

# Indicates how long it usually takes for the underlying storage to propagate all writes
# to be available to subsequent reads.
STORAGE_CONSISTENCY_DELAY_MILLIS = 3000


class RequestQueue(BaseStorage):
    """Represents a queue of URLs to crawl.

    Can be used for deep crawling of websites where you start with several URLs and then recursively
    follow links to other pages. The data structure supports both breadth-first and depth-first crawling orders.

    Each URL is represented using an instance of the {@apilink Request} class.
    The queue can only contain unique URLs. More precisely, it can only contain request dictionaries
    with distinct `uniqueKey` properties. By default, `uniqueKey` is generated from the URL, but it can also be overridden.
    To add a single URL multiple times to the queue,
    corresponding request dictionary will need to have different `uniqueKey` properties.

    Do not instantiate this class directly, use the `Actor.open_request_queue()` function instead.

    `RequestQueue` stores its data either on local disk or in the Apify cloud,
    depending on whether the `APIFY_LOCAL_STORAGE_DIR` or `APIFY_TOKEN` environment variables are set.

    If the `APIFY_LOCAL_STORAGE_DIR` environment variable is set, the data is stored in
    the local directory in the following files:
    ```
    {APIFY_LOCAL_STORAGE_DIR}/request_queues/{QUEUE_ID}/{REQUEST_ID}.json
    ```
    Note that `{QUEUE_ID}` is the name or ID of the request queue. The default request queue has ID: `default`,
    unless you override it by setting the `APIFY_DEFAULT_REQUEST_QUEUE_ID` environment variable.
    The `{REQUEST_ID}` is the id of the request.

    If the `APIFY_TOKEN` environment variable is set but `APIFY_LOCAL_STORAGE_DIR` is not, the data is stored in the
    [Apify Request Queue](https://docs.apify.com/storage/request-queue)
    cloud storage.
    """

    _request_queue_client: Union[RequestQueueClientAsync, RequestQueueClient]
    _client_key = _crypto_random_object_id()
    _queue_head_dict: OrderedDictType[str, str]
    _query_queue_head_promise: Optional[Coroutine]
    _in_progress: Set[str]
    _last_activity: datetime
    _internal_timeout_seconds = 5 * 60
    _recently_handled: LRUCache[bool]
    _assumed_total_count = 0
    _assumed_handled_count = 0
    _requests_cache: LRUCache[Dict]

    @ignore_docs
    def __init__(self, id: str, name: Optional[str], client: Union[ApifyClientAsync, MemoryStorageClient], config: Configuration) -> None:
        """Create a `RequestQueue` instance.

        Do not use the constructor directly, use the `Actor.open_request_queue()` function instead.

        Args:
            id (str): ID of the request queue.
            name (str, optional): Name of the request queue.
            client (ApifyClientAsync or MemoryStorageClient): The storage client which should be used.
            config (Configuration): The configuration which should be used.
        """
        super().__init__(id=id, name=name, client=client, config=config)

        self._request_queue_client = client.request_queue(self._id, client_key=self._client_key)
        self._queue_head_dict = OrderedDict()
        self._query_queue_head_promise = None
        self._in_progress = set()
        self._last_activity = datetime.now(timezone.utc)
        self._recently_handled = LRUCache[bool](max_length=RECENTLY_HANDLED_CACHE_SIZE)
        self._requests_cache = LRUCache(max_length=MAX_CACHED_REQUESTS)

    @classmethod
    def _get_human_friendly_label(cls) -> str:
        return 'Request queue'

    @classmethod
    def _get_default_id(cls, config: Configuration) -> str:
        return config.default_request_queue_id

    @classmethod
    def _get_single_storage_client(
        cls,
        id: str,
        client: Union[ApifyClientAsync, MemoryStorageClient],
    ) -> Union[RequestQueueClientAsync, RequestQueueClient]:
        return client.request_queue(id)

    @classmethod
    def _get_storage_collection_client(
        cls,
        client: Union[ApifyClientAsync, MemoryStorageClient],
    ) -> Union[RequestQueueCollectionClientAsync, RequestQueueCollectionClient]:
        return client.request_queues()

    async def add_request(self, request: Dict, *, forefront: bool = False) -> Dict:
        """Add a request to the queue.

        Args:
            request (dict): The request to add to the queue
            forefront (bool, optional): Whether to add the request to the head or the end of the queue

        Returns:
            dict: Information about the queue operation with keys `requestId`, `uniqueKey`, `wasAlreadyPresent`, `wasAlreadyHandled`.
        """
        _budget_ow(request, {
            'url': (str, True),
        })
        self._last_activity = datetime.now(timezone.utc)

        if request.get('uniqueKey') is None:
            request['uniqueKey'] = request['url']  # TODO: Check Request class in crawlee and replicate uniqueKey generation logic...

        cache_key = _unique_key_to_request_id(request['uniqueKey'])
        cached_info = self._requests_cache.get(cache_key)

        if cached_info:
            request['id'] = cached_info['id']
            return {
                'wasAlreadyPresent': True,
                # We may assume that if request is in local cache then also the information if the
                # request was already handled is there because just one client should be using one queue.
                'wasAlreadyHandled': cached_info['isHandled'],
                'requestId': cached_info['id'],
                'uniqueKey': cached_info['uniqueKey'],
            }

        queue_operation_info = await self._request_queue_client.add_request(request, forefront=forefront)
        queue_operation_info['uniqueKey'] = request['uniqueKey']

        self._cache_request(cache_key, queue_operation_info)

        request_id, was_already_present = queue_operation_info['requestId'], queue_operation_info['wasAlreadyPresent']
        if not was_already_present and request_id not in self._in_progress and self._recently_handled.get(request_id) is None:
            self._assumed_total_count += 1

            self._maybe_add_request_to_queue_head(request_id, forefront)

        return queue_operation_info

    async def get_request(self, request_id: str) -> Optional[Dict]:
        """Retrieve a request from the queue.

        Args:
            request_id (str): ID of the request to retrieve.

        Returns:
            dict, optional: The retrieved request, or `None`, if it does not exist.
        """
        _budget_ow(request_id, (str, True), 'request_id')
        return await self._request_queue_client.get_request(request_id)

    async def fetch_next_request(self) -> Optional[Dict]:
        """Return the next request in the queue to be processed.

        Once you successfully finish processing of the request, you need to call
        `RequestQueue.mark_request_as_handled` to mark the request as handled in the queue.
        If there was some error in processing the request, call `RequestQueue.reclaim_request` instead,
        so that the queue will give the request to some other consumer in another call to the `fetch_next_request` method.

        Note that the `None` return value does not mean the queue processing finished, it means there are currently no pending requests.
        To check whether all requests in queue were finished, use `RequestQueue.is_finished` instead.

        Returns:
            dict, optional: The request or `None` if there are no more pending requests.
        """
        await self._ensure_head_is_non_empty()

        # We are likely done at this point.
        if len(self._queue_head_dict) == 0:
            return None

        next_request_id, _ = self._queue_head_dict.popitem(last=False)  # ~removeFirst()

        # This should never happen, but...
        if next_request_id in self._in_progress or self._recently_handled.get(next_request_id):
            logger.warning('Queue head returned a request that is already in progress?!', extra={
                'nextRequestId': next_request_id,
                'inProgress': next_request_id in self._in_progress,
                'recentlyHandled': next_request_id in self._recently_handled,
            })
            return None
        self._in_progress.add(next_request_id)
        self._last_activity = datetime.now(timezone.utc)

        try:
            request = await self.get_request(next_request_id)
        except Exception as e:
            # On error, remove the request from in progress, otherwise it would be there forever
            self._in_progress.remove(next_request_id)
            raise e

        # NOTE: It can happen that the queue head index is inconsistent with the main queue table. This can occur in two situations:

        """ 1) Queue head index is ahead of the main table and the request is not present in the main table yet (i.e. getRequest() returned null).
                In this case, keep the request marked as in progress for a short while,
                so that isFinished() doesn't return true and _ensureHeadIsNonEmpty() doesn't not load the request
                into the queueHeadDict straight again. After the interval expires, fetchNextRequest()
                will try to fetch this request again, until it eventually appears in the main table.
        """
        if request is None:
            logger.debug('Cannot find a request from the beginning of queue, will be retried later', extra={'nextRequestId': next_request_id})
            asyncio.get_running_loop().call_later(STORAGE_CONSISTENCY_DELAY_MILLIS // 1000, lambda: self._in_progress.remove(next_request_id))
            return None

        """ 2) Queue head index is behind the main table and the underlying request was already handled
               (by some other client, since we keep the track of handled requests in recentlyHandled dictionary).
               We just add the request to the recentlyHandled dictionary so that next call to _ensureHeadIsNonEmpty()
               will not put the request again to queueHeadDict.
        """
        if request.get('handledAt') is not None:
            logger.debug('Request fetched from the beginning of queue was already handled', extra={'nextRequestId': next_request_id})
            self._recently_handled[next_request_id] = True
            return None

        return request

    async def mark_request_as_handled(self, request: Dict) -> Optional[Dict]:
        """Mark a request as handled after successful processing.

        Handled requests will never again be returned by the `RequestQueue.fetch_next_request` method.

        Args:
            request (dict): The request to mark as handled.

        Returns:
            dict, optional: Information about the queue operation with keys `requestId`, `uniqueKey`, `wasAlreadyPresent`, `wasAlreadyHandled`.
                `None` if the given request was not in progress.
        """
        _budget_ow(request, {
            'id': (str, True),
            'uniqueKey': (str, True),
            'handledAt': (datetime, False),
        })
        self._last_activity = datetime.now(timezone.utc)
        if request['id'] not in self._in_progress:
            logger.debug('Cannot mark request as handled, because it is not in progress!', extra={'requestId': request['id']})
            return None

        request['handledAt'] = request.get('handledAt', datetime.now(timezone.utc))
        queue_operation_info = await self._request_queue_client.update_request({**request})
        queue_operation_info['uniqueKey'] = request['uniqueKey']

        self._in_progress.remove(request['id'])
        self._recently_handled[request['id']] = True

        if not queue_operation_info['wasAlreadyHandled']:
            self._assumed_handled_count += 1

        self._cache_request(_unique_key_to_request_id(request['uniqueKey']), queue_operation_info)

        return queue_operation_info

    async def reclaim_request(self, request: Dict, forefront: bool = False) -> Optional[Dict]:
        """Reclaim a failed request back to the queue.

        The request will be returned for processing later again
        by another call to `RequestQueue.fetchNextRequest`.

        Args:
            request (dict): The request to return to the queue.
            forefront (bool, optional): Whether to add the request to the head or the end of the queue
        Returns:
            dict, optional: Information about the queue operation with keys `requestId`, `uniqueKey`, `wasAlreadyPresent`, `wasAlreadyHandled`.
                `None` if the given request was not in progress.
        """
        _budget_ow(request, {
            'id': (str, True),
            'uniqueKey': (str, True),
        })
        self._last_activity = datetime.now(timezone.utc)

        if request['id'] not in self._in_progress:
            logger.debug('Cannot reclaim request, because it is not in progress!', extra={'requestId': request['id']})
            return None

        # TODO: If request hasn't been changed since the last getRequest(),
        #       we don't need to call updateRequest() and thus improve performance.
        queue_operation_info = await self._request_queue_client.update_request(request, forefront=forefront)
        queue_operation_info['uniqueKey'] = request['uniqueKey']
        self._cache_request(_unique_key_to_request_id(request['uniqueKey']), queue_operation_info)

        # Wait a little to increase a chance that the next call to fetchNextRequest() will return the request with updated data.
        # This is to compensate for the limitation of DynamoDB, where writes might not be immediately visible to subsequent reads.
        def callback() -> None:
            if request['id'] not in self._in_progress:
                logger.debug('The request is no longer marked as in progress in the queue?!', {'requestId': request['id']})
                return

            self._in_progress.remove(request['id'])

            # Performance optimization: add request straight to head if possible
            self._maybe_add_request_to_queue_head(request['id'], forefront)

        asyncio.get_running_loop().call_later(STORAGE_CONSISTENCY_DELAY_MILLIS // 1000, callback)

        return queue_operation_info

    def _in_progress_count(self) -> int:
        return len(self._in_progress)

    async def is_empty(self) -> bool:
        """Check whether the queue is empty.

        Returns:
            bool: `True` if the next call to `RequestQueue.fetchNextRequest` would return `None`, otherwise `False`.
        """
        await self._ensure_head_is_non_empty()
        return len(self._queue_head_dict) == 0

    async def is_finished(self) -> bool:
        """Check whether the queue is finished.

        Due to the nature of distributed storage used by the queue,
        the function might occasionally return a false negative,
        but it will never return a false positive.

        Returns:
            bool: `True` if all requests were already handled and there are no more left. `False` otherwise.
        """
        seconds_since_last_activity = (datetime.now(timezone.utc) - self._last_activity).seconds
        if self._in_progress_count() > 0 and seconds_since_last_activity > self._internal_timeout_seconds:
            message = f'The request queue seems to be stuck for {self._internal_timeout_seconds}s, resetting internal state.'
            logger.warning(message)
            self._reset()

        if (len(self._queue_head_dict) > 0 or self._in_progress_count() > 0):
            return False

        is_head_consistent = await self._ensure_head_is_non_empty(True)
        return is_head_consistent and len(self._queue_head_dict) == 0 and self._in_progress_count() == 0

    def _reset(self) -> None:
        self._queue_head_dict.clear()
        self._query_queue_head_promise = None
        self._in_progress.clear()
        self._recently_handled.clear()
        self._assumed_total_count = 0
        self._assumed_handled_count = 0
        self._requests_cache.clear()
        self._last_activity = datetime.now(timezone.utc)

    def _cache_request(self, cache_key: str, queue_operation_info: Dict) -> None:
        self._requests_cache[cache_key] = {
            'id': queue_operation_info['requestId'],
            'isHandled': queue_operation_info['wasAlreadyHandled'],
            'uniqueKey': queue_operation_info['uniqueKey'],
            'wasAlreadyHandled': queue_operation_info['wasAlreadyHandled'],
        }

    async def _queue_query_head(self, limit: int) -> Dict:
        query_started_at = datetime.now(timezone.utc)

        list_head = await self._request_queue_client.list_head(limit=limit)
        for request in list_head['items']:
            # Queue head index might be behind the main table, so ensure we don't recycle requests
            if not request['id'] or not request['uniqueKey'] or request['id'] in self._in_progress or self._recently_handled.get(request['id']):
                continue
            self._queue_head_dict[request['id']] = request['id']
            self._cache_request(_unique_key_to_request_id(request['uniqueKey']), {
                'requestId': request['id'],
                'wasAlreadyHandled': False,
                'wasAlreadyPresent': True,
                'uniqueKey': request['uniqueKey'],
            })

        # This is needed so that the next call to _ensureHeadIsNonEmpty() will fetch the queue head again.
        self._query_queue_head_promise = None

        return {
            'wasLimitReached': len(list_head['items']) >= limit,
            'prevLimit': limit,
            'queueModifiedAt': list_head['queueModifiedAt'],
            'queryStartedAt': query_started_at,
            'hadMultipleClients': list_head['hadMultipleClients'],
        }

    async def _ensure_head_is_non_empty(self, ensure_consistency: bool = False, limit: Optional[int] = None, iteration: int = 0) -> bool:
        # If is nonempty resolve immediately.
        if len(self._queue_head_dict) > 0:
            return True

        if limit is None:
            limit = max(self._in_progress_count() * QUERY_HEAD_BUFFER, QUERY_HEAD_MIN_LENGTH)

        if self._query_queue_head_promise is None:
            self._query_queue_head_promise = self._queue_query_head(limit)

        queue_head = await self._query_queue_head_promise

        # TODO: I feel this code below can be greatly simplified... (comes from TS implementation *wink*)

        """ If queue is still empty then one of the following holds:
        - the other calls waiting for this promise already consumed all the returned requests
        - the limit was too low and contained only requests in progress
        - the writes from other clients were not propagated yet
        - the whole queue was processed and we are done
        """

        # If limit was not reached in the call then there are no more requests to be returned.
        if (queue_head['prevLimit'] >= REQUEST_QUEUE_HEAD_MAX_LIMIT):
            logger.warning('Reached the maximum number of requests in progress', extra={'limit': REQUEST_QUEUE_HEAD_MAX_LIMIT})

        should_repeat_with_higher_limit = len(
            self._queue_head_dict) == 0 and queue_head['wasLimitReached'] and queue_head['prevLimit'] < REQUEST_QUEUE_HEAD_MAX_LIMIT

        # If ensureConsistency=true then we must ensure that either:
        # - queueModifiedAt is older than queryStartedAt by at least API_PROCESSED_REQUESTS_DELAY_MILLIS
        # - hadMultipleClients=false and this.assumedTotalCount<=this.assumedHandledCount
        is_database_consistent = (queue_head['queryStartedAt'] - queue_head['queueModifiedAt'].replace(tzinfo=timezone.utc)
                                  ).seconds >= (API_PROCESSED_REQUESTS_DELAY_MILLIS // 1000)
        is_locally_consistent = not queue_head['hadMultipleClients'] and self._assumed_total_count <= self._assumed_handled_count
        # Consistent information from one source is enough to consider request queue finished.
        should_repeat_for_consistency = ensure_consistency and not is_database_consistent and not is_locally_consistent

        # If both are false then head is consistent and we may exit.
        if not should_repeat_with_higher_limit and not should_repeat_for_consistency:
            return True

        # If we are querying for consistency then we limit the number of queries to MAX_QUERIES_FOR_CONSISTENCY.
        # If this is reached then we return false so that empty() and finished() returns possibly false negative.
        if not should_repeat_with_higher_limit and iteration > MAX_QUERIES_FOR_CONSISTENCY:
            return False

        next_limit = round(queue_head['prevLimit'] * 1.5) if should_repeat_with_higher_limit else queue_head['prevLimit']

        # If we are repeating for consistency then wait required time.
        if should_repeat_for_consistency:
            delay_seconds = (API_PROCESSED_REQUESTS_DELAY_MILLIS // 1000) - \
                (datetime.now(timezone.utc) - queue_head['queueModifiedAt']).seconds
            logger.info(f'Waiting for {delay_seconds}s before considering the queue as finished to ensure that the data is consistent.')
            await asyncio.sleep(delay_seconds)

        return await self._ensure_head_is_non_empty(ensure_consistency, next_limit, iteration + 1)

    def _maybe_add_request_to_queue_head(self, request_id: str, forefront: bool) -> None:
        if forefront:
            self._queue_head_dict[request_id] = request_id
            # Move to start, i.e. forefront of the queue
            self._queue_head_dict.move_to_end(request_id, last=False)
        elif self._assumed_total_count < QUERY_HEAD_MIN_LENGTH:
            # OrderedDict puts the item to the end of the queue by default
            self._queue_head_dict[request_id] = request_id

    async def drop(self) -> None:
        """Remove the request queue either from the Apify cloud storage or from the local directory."""
        await self._request_queue_client.delete()
        self._remove_from_cache()

    async def get_info(self) -> Optional[Dict]:
        """Get an object containing general information about the request queue.

        Returns:
            dict: Object returned by calling the GET request queue API endpoint.
        """
        return await self._request_queue_client.get()

    @classmethod
    async def open(
        cls,
        *,
        id: Optional[str] = None,
        name: Optional[str] = None,
        force_cloud: bool = False,
        config: Optional[Configuration] = None,
    ) -> 'RequestQueue':
        """Open a request queue.

        Request queue represents a queue of URLs to crawl, which is stored either on local filesystem or in the Apify cloud.
        The queue is used for deep crawling of websites, where you start with several URLs and then
        recursively follow links to other pages. The data structure supports both breadth-first
        and depth-first crawling orders.

        Args:
            id (str, optional): ID of the request queue to be opened.
                If neither `id` nor `name` are provided, the method returns the default request queue associated with the actor run.
                If the request queue with the given ID does not exist, it raises an error.
            name (str, optional): Name of the request queue to be opened.
                If neither `id` nor `name` are provided, the method returns the default request queue associated with the actor run.
                If the request queue with the given name does not exist, it is created.
            force_cloud (bool, optional): If set to True, it will open a request queue on the Apify Platform even when running the actor locally.
                Defaults to False.
            config (Configuration, optional): A `Configuration` instance, uses global configuration if omitted.

        Returns:
            RequestQueue: An instance of the `RequestQueue` class for the given ID or name.
        """
        return await super().open(id=id, name=name, force_cloud=force_cloud, config=config)



import asyncio
from abc import ABC, abstractmethod
from typing import Dict, Generic, Optional, TypeVar, Union, cast

from typing_extensions import Self

from apify_client import ApifyClientAsync

from .._memory_storage import MemoryStorageClient
from .._memory_storage.resource_clients import BaseResourceClient, BaseResourceCollectionClient
from .._utils import ignore_docs
from ..config import Configuration
from .storage_client_manager import StorageClientManager

BaseResourceClientType = TypeVar('BaseResourceClientType', bound=BaseResourceClient)
BaseResourceCollectionClientType = TypeVar('BaseResourceCollectionClientType', bound=BaseResourceCollectionClient)


@ignore_docs
class BaseStorage(ABC, Generic[BaseResourceClientType, BaseResourceCollectionClientType]):
    """A class for managing storages."""

    _id: str
    _name: Optional[str]
    _storage_client: Union[ApifyClientAsync, MemoryStorageClient]
    _config: Configuration

    _cache_by_id: Optional[Dict[str, Self]] = None
    _cache_by_name: Optional[Dict[str, Self]] = None
    _storage_creating_lock: Optional[asyncio.Lock] = None

    def __init__(self, id: str, name: Optional[str], client: Union[ApifyClientAsync, MemoryStorageClient], config: Configuration):
        """Initialize the storage.

        Do not use this method directly, but use `Actor.open_<STORAGE>()` instead.

        Args:
            id (str): The storage id
            name (str, optional): The storage name
            client (ApifyClientAsync or MemoryStorageClient): The storage client
            config (Configuration): The configuration
        """
        self._id = id
        self._name = name
        self._storage_client = client
        self._config = config

    @classmethod
    @abstractmethod
    def _get_human_friendly_label(cls) -> str:
        raise NotImplementedError('You must override this method in the subclass!')

    @classmethod
    @abstractmethod
    def _get_default_id(cls, config: Configuration) -> str:
        raise NotImplementedError('You must override this method in the subclass!')

    @classmethod
    @abstractmethod
    def _get_single_storage_client(cls, id: str, client: Union[ApifyClientAsync, MemoryStorageClient]) -> BaseResourceClientType:
        raise NotImplementedError('You must override this method in the subclass!')

    @classmethod
    @abstractmethod
    def _get_storage_collection_client(cls, client: Union[ApifyClientAsync, MemoryStorageClient]) -> BaseResourceCollectionClientType:
        raise NotImplementedError('You must override this method in the subclass!')

    @classmethod
    def _ensure_class_initialized(cls) -> None:
        if cls._cache_by_id is None:
            cls._cache_by_id = {}
        if cls._cache_by_name is None:
            cls._cache_by_name = {}
        if cls._storage_creating_lock is None:
            cls._storage_creating_lock = asyncio.Lock()

    @classmethod
    @abstractmethod
    async def open(
        cls,
        *,
        id: Optional[str] = None,
        name: Optional[str] = None,
        force_cloud: bool = False,
        config: Optional[Configuration] = None,
    ) -> Self:
        """Open a storage, or return a cached storage object if it was opened before.

        Opens a storage with the given ID or name.
        Returns the cached storage object if the storage was opened before.

        Args:
            id (str, optional): ID of the storage to be opened.
                If neither `id` nor `name` are provided, the method returns the default storage associated with the actor run.
                If the storage with the given ID does not exist, it raises an error.
            name (str, optional): Name of the storage to be opened.
                If neither `id` nor `name` are provided, the method returns the default storage associated with the actor run.
                If the storage with the given name does not exist, it is created.
            force_cloud (bool, optional): If set to True, it will open a storage on the Apify Platform even when running the actor locally.
                Defaults to False.
            config (Configuration, optional): A `Configuration` instance, uses global configuration if omitted.

        Returns:
            An instance of the storage.
        """
        cls._ensure_class_initialized()
        assert cls._cache_by_id is not None
        assert cls._cache_by_name is not None

        assert not (id and name)

        used_config = config or Configuration.get_global_configuration()
        used_client = StorageClientManager.get_storage_client(force_cloud=force_cloud)

        is_default_storage_on_local = False
        # Fetch default ID if no ID or name was passed
        if not id and not name:
            if isinstance(used_client, MemoryStorageClient):
                is_default_storage_on_local = True
            id = cls._get_default_id(used_config)

        # Try to get the storage instance from cache
        cached_storage = None
        if id:
            cached_storage = cls._cache_by_id.get(id)
        elif name:
            cached_storage = cls._cache_by_name.get(name)

        if cached_storage is not None:
            # This cast is needed since MyPy doesn't understand very well that Self and Storage are the same
            return cast(Self, cached_storage)

        # Purge default storages if configured
        if used_config.purge_on_start:
            if isinstance(used_client, MemoryStorageClient):
                await used_client._purge_on_start()

        assert cls._storage_creating_lock is not None
        async with cls._storage_creating_lock:
            # Create the storage
            if id and not is_default_storage_on_local:
                single_storage_client = cls._get_single_storage_client(id, used_client)
                storage_info = await single_storage_client.get()
                if not storage_info:
                    storage_label = cls._get_human_friendly_label()
                    raise RuntimeError(f'{storage_label} with id "{id}" does not exist!')
            elif is_default_storage_on_local:
                storage_collection_client = cls._get_storage_collection_client(used_client)
                storage_info = await storage_collection_client.get_or_create(name=name, _id=id)
            else:
                storage_collection_client = cls._get_storage_collection_client(used_client)
                storage_info = await storage_collection_client.get_or_create(name=name)

            storage = cls(storage_info['id'], storage_info.get('name'), used_client, used_config)

            # Cache by id and name
            cls._cache_by_id[storage._id] = storage
            if storage._name is not None:
                cls._cache_by_name[storage._name] = storage

        return storage

    def _remove_from_cache(self) -> None:
        if self.__class__._cache_by_id is not None:
            del self.__class__._cache_by_id[self._id]

        if self._name and self.__class__._cache_by_name is not None:
            del self.__class__._cache_by_name[self._name]



from typing import Dict, List, Optional, Type

from apify_client._utils import ListPage

from ..._utils import ignore_docs
from .base_resource_collection_client import BaseResourceCollectionClient
from .request_queue import RequestQueueClient


@ignore_docs
class RequestQueueCollectionClient(BaseResourceCollectionClient):
    """Sub-client for manipulating request queues."""

    def _get_storage_client_cache(self) -> List[RequestQueueClient]:
        return self._memory_storage_client._request_queues_handled

    def _get_resource_client_class(self) -> Type[RequestQueueClient]:
        return RequestQueueClient

    async def list(self) -> ListPage:
        """List the available request queues.

        Returns:
            ListPage: The list of available request queues matching the specified filters.
        """
        return await super().list()

    async def get_or_create(
        self,
        *,
        name: Optional[str] = None,
        schema: Optional[Dict] = None,
        _id: Optional[str] = None,
    ) -> Dict:
        """Retrieve a named request queue, or create a new one when it doesn't exist.

        Args:
            name (str, optional): The name of the request queue to retrieve or create.
            schema (Dict, optional): The schema of the request queue

        Returns:
            dict: The retrieved or newly-created request queue.
        """
        return await super().get_or_create(name=name, schema=schema, _id=_id)



from typing import Dict, List, Optional, Type

from apify_client._utils import ListPage

from ..._utils import ignore_docs
from .base_resource_collection_client import BaseResourceCollectionClient
from .key_value_store import KeyValueStoreClient


@ignore_docs
class KeyValueStoreCollectionClient(BaseResourceCollectionClient):
    """Sub-client for manipulating key-value stores."""

    def _get_storage_client_cache(self) -> List[KeyValueStoreClient]:
        return self._memory_storage_client._key_value_stores_handled

    def _get_resource_client_class(self) -> Type[KeyValueStoreClient]:
        return KeyValueStoreClient

    async def list(self) -> ListPage:
        """List the available key-value stores.

        Returns:
            ListPage: The list of available key-value stores matching the specified filters.
        """
        return await super().list()

    async def get_or_create(
        self,
        *,
        name: Optional[str] = None,
        schema: Optional[Dict] = None,
        _id: Optional[str] = None,
    ) -> Dict:
        """Retrieve a named key-value store, or create a new one when it doesn't exist.

        Args:
            name (str, optional): The name of the key-value store to retrieve or create.
            schema (Dict, optional): The schema of the key-value store

        Returns:
            dict: The retrieved or newly-created key-value store.
        """
        return await super().get_or_create(name=name, schema=schema, _id=_id)



from abc import ABC, abstractmethod
from operator import itemgetter
from typing import TYPE_CHECKING, Dict, Generic, List, Optional, Type, TypeVar

from apify_client._utils import ListPage

from ..._utils import ignore_docs
from ..file_storage_utils import _update_metadata
from .base_resource_client import BaseResourceClient

if TYPE_CHECKING:
    from ..memory_storage_client import MemoryStorageClient


ResourceClientType = TypeVar('ResourceClientType', bound=BaseResourceClient, contravariant=True)


@ignore_docs
class BaseResourceCollectionClient(ABC, Generic[ResourceClientType]):
    """Base class for resource collection clients."""

    _base_storage_directory: str
    _memory_storage_client: 'MemoryStorageClient'

    def __init__(self, *, base_storage_directory: str, memory_storage_client: 'MemoryStorageClient') -> None:
        """Initialize the DatasetCollectionClient with the passed arguments."""
        self._base_storage_directory = base_storage_directory
        self._memory_storage_client = memory_storage_client

    @abstractmethod
    def _get_storage_client_cache(self) -> List[ResourceClientType]:
        raise NotImplementedError('You must override this method in the subclass!')

    @abstractmethod
    def _get_resource_client_class(self) -> Type[ResourceClientType]:
        raise NotImplementedError('You must override this method in the subclass!')

    @abstractmethod
    async def list(self) -> ListPage:
        """List the available storages.

        Returns:
            ListPage: The list of available storages matching the specified filters.
        """
        storage_client_cache = self._get_storage_client_cache()

        items = [storage._to_resource_info() for storage in storage_client_cache]

        return ListPage({
            'total': len(items),
            'count': len(items),
            'offset': 0,
            'limit': len(items),
            'desc': False,
            'items': sorted(items, key=itemgetter('createdAt')),
        })

    @abstractmethod
    async def get_or_create(
        self,
        *,
        name: Optional[str] = None,
        schema: Optional[Dict] = None,  # noqa: U100
        _id: Optional[str] = None,
    ) -> Dict:
        """Retrieve a named storage, or create a new one when it doesn't exist.

        Args:
            name (str, optional): The name of the storage to retrieve or create.
            schema (Dict, optional): The schema of the storage

        Returns:
            dict: The retrieved or newly-created storage.
        """
        resource_client_class = self._get_resource_client_class()
        storage_client_cache = self._get_storage_client_cache()

        if name or _id:
            found = resource_client_class._find_or_create_client_by_id_or_name(memory_storage_client=self._memory_storage_client, name=name, id=_id)
            if found:
                return found._to_resource_info()

        new_resource = resource_client_class(
            id=_id,
            name=name,
            base_storage_directory=self._base_storage_directory,
            memory_storage_client=self._memory_storage_client,
        )
        storage_client_cache.append(new_resource)

        resource_info = new_resource._to_resource_info()

        # Write to the disk
        await _update_metadata(
            data=resource_info,
            entity_directory=new_resource._resource_directory,
            write_metadata=self._memory_storage_client._write_metadata,
        )

        return resource_info



import csv
import io
import math
from typing import AsyncIterator, Dict, Iterable, Iterator, List, Optional, Union

from apify_client import ApifyClientAsync
from apify_client._utils import ListPage
from apify_client.clients import DatasetClientAsync, DatasetCollectionClientAsync

from .._memory_storage import MemoryStorageClient
from .._memory_storage.resource_clients import DatasetClient, DatasetCollectionClient
from .._types import JSONSerializable
from .._utils import _json_dumps, _wrap_internal, ignore_docs
from ..config import Configuration
from ..consts import MAX_PAYLOAD_SIZE_BYTES
from .base_storage import BaseStorage
from .key_value_store import KeyValueStore

SAFETY_BUFFER_PERCENT = 0.01 / 100  # 0.01%
EFFECTIVE_LIMIT_BYTES = MAX_PAYLOAD_SIZE_BYTES - math.ceil(MAX_PAYLOAD_SIZE_BYTES * SAFETY_BUFFER_PERCENT)


def _check_and_serialize(item: JSONSerializable, index: Optional[int] = None) -> str:
    """Accept a JSON serializable object as an input, validate its serializability and its serialized size against `EFFECTIVE_LIMIT_BYTES`."""
    s = ' ' if index is None else f' at index {index} '

    try:
        payload = _json_dumps(item)
    except Exception as e:
        raise ValueError(f'Data item{s}is not serializable to JSON.') from e

    length_bytes = len(payload.encode('utf-8'))
    if length_bytes > EFFECTIVE_LIMIT_BYTES:
        raise ValueError(f'Data item{s}is too large (size: {length_bytes} bytes, limit: {EFFECTIVE_LIMIT_BYTES} bytes)')

    return payload


def _chunk_by_size(items: Iterable[str]) -> Iterator[str]:
    """Take an array of JSONs, produce iterator of chunked JSON arrays respecting `EFFECTIVE_LIMIT_BYTES`.

    Takes an array of JSONs (payloads) as input and produces an iterator of JSON strings
    where each string is a JSON array of payloads with a maximum size of `EFFECTIVE_LIMIT_BYTES` per one
    JSON array. Fits as many payloads as possible into a single JSON array and then moves
    on to the next, preserving item order.

    The function assumes that none of the items is larger than `EFFECTIVE_LIMIT_BYTES` and does not validate.
    """
    last_chunk_bytes = 2  # Add 2 bytes for [] wrapper.
    current_chunk = []

    for payload in items:
        length_bytes = len(payload.encode('utf-8'))

        if last_chunk_bytes + length_bytes <= EFFECTIVE_LIMIT_BYTES:
            current_chunk.append(payload)
            last_chunk_bytes += length_bytes + 1  # Add 1 byte for ',' separator.
        else:
            yield f'[{",".join(current_chunk)}]'
            current_chunk = [payload]
            last_chunk_bytes = length_bytes + 2  # Add 2 bytes for [] wrapper.

    yield f'[{",".join(current_chunk)}]'


class Dataset(BaseStorage):
    """The `Dataset` class represents a store for structured data where each object stored has the same attributes.

    You can imagine it as a table, where each object is a row and its attributes are columns.
    Dataset is an append-only storage - you can only add new records to it but you cannot modify or remove existing records.
    Typically it is used to store crawling results.

    Do not instantiate this class directly, use the `Actor.open_dataset()` function instead.

    `Dataset` stores its data either on local disk or in the Apify cloud,
    depending on whether the `APIFY_LOCAL_STORAGE_DIR` or `APIFY_TOKEN` environment variables are set.

    If the `APIFY_LOCAL_STORAGE_DIR` environment variable is set, the data is stored in
    the local directory in the following files:
    ```
    {APIFY_LOCAL_STORAGE_DIR}/datasets/{DATASET_ID}/{INDEX}.json
    ```
    Note that `{DATASET_ID}` is the name or ID of the dataset. The default dataset has ID: `default`,
    unless you override it by setting the `APIFY_DEFAULT_DATASET_ID` environment variable.
    Each dataset item is stored as a separate JSON file, where `{INDEX}` is a zero-based index of the item in the dataset.

    If the `APIFY_TOKEN` environment variable is set but `APIFY_LOCAL_STORAGE_DIR` is not, the data is stored in the
    [Apify Dataset](https://docs.apify.com/storage/dataset) cloud storage.
    """

    _id: str
    _name: Optional[str]
    _dataset_client: Union[DatasetClientAsync, DatasetClient]

    @ignore_docs
    def __init__(self, id: str, name: Optional[str], client: Union[ApifyClientAsync, MemoryStorageClient], config: Configuration) -> None:
        """Create a `Dataset` instance.

        Do not use the constructor directly, use the `Actor.open_dataset()` function instead.

        Args:
            id (str): ID of the dataset.
            name (str, optional): Name of the dataset.
            client (ApifyClientAsync or MemoryStorageClient): The storage client which should be used.
            config (Configuration): The configuration which should be used.
        """
        super().__init__(id=id, name=name, client=client, config=config)

        self.get_data = _wrap_internal(self._get_data_internal, self.get_data)  # type: ignore
        self.push_data = _wrap_internal(self._push_data_internal, self.push_data)  # type: ignore
        self.export_to_json = _wrap_internal(self._export_to_json_internal, self.export_to_json)  # type: ignore
        self.export_to_csv = _wrap_internal(self._export_to_csv_internal, self.export_to_csv)  # type: ignore

        self._dataset_client = client.dataset(self._id)

    @classmethod
    def _get_human_friendly_label(cls) -> str:
        return 'Dataset'

    @classmethod
    def _get_default_id(cls, config: Configuration) -> str:
        return config.default_dataset_id

    @classmethod
    def _get_single_storage_client(cls, id: str, client: Union[ApifyClientAsync, MemoryStorageClient]) -> Union[DatasetClientAsync, DatasetClient]:
        return client.dataset(id)

    @classmethod
    def _get_storage_collection_client(
        cls,
        client: Union[ApifyClientAsync, MemoryStorageClient],
    ) -> Union[DatasetCollectionClientAsync, DatasetCollectionClient]:
        return client.datasets()

    @classmethod
    async def push_data(cls, data: JSONSerializable) -> None:
        """Store an object or an array of objects to the dataset.

        The size of the data is limited by the receiving API and therefore `push_data()` will only
        allow objects whose JSON representation is smaller than 9MB. When an array is passed,
        none of the included objects may be larger than 9MB, but the array itself may be of any size.

        Args:
            data (JSONSerializable): dict or array of dicts containing data to be stored in the default dataset.
                The JSON representation of each item must be smaller than 9MB.
        """
        dataset = await cls.open()
        return await dataset.push_data(data)

    async def _push_data_internal(self, data: JSONSerializable) -> None:
        # Handle singular items
        if not isinstance(data, list):
            payload = _check_and_serialize(data)
            return await self._dataset_client.push_items(payload)

        # Handle lists
        payloads_generator = (_check_and_serialize(item, index) for index, item in enumerate(data))

        # Invoke client in series to preserve the order of data
        for chunk in _chunk_by_size(payloads_generator):
            await self._dataset_client.push_items(chunk)

    @classmethod
    async def get_data(
        cls,
        *,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        clean: Optional[bool] = None,
        desc: Optional[bool] = None,
        fields: Optional[List[str]] = None,
        omit: Optional[List[str]] = None,
        unwind: Optional[str] = None,
        skip_empty: Optional[bool] = None,
        skip_hidden: Optional[bool] = None,
        flatten: Optional[List[str]] = None,
        view: Optional[str] = None,
    ) -> ListPage:
        """Get items from the dataset.

        Args:
            offset (int, optional): Number of items that should be skipped at the start. The default value is 0
            limit (int, optional): Maximum number of items to return. By default there is no limit.
            desc (bool, optional): By default, results are returned in the same order as they were stored.
                To reverse the order, set this parameter to True.
            clean (bool, optional): If True, returns only non-empty items and skips hidden fields (i.e. fields starting with the # character).
                The clean parameter is just a shortcut for skip_hidden=True and skip_empty=True parameters.
                Note that since some objects might be skipped from the output, that the result might contain less items than the limit value.
            fields (list of str, optional): A list of fields which should be picked from the items,
                only these fields will remain in the resulting record objects.
                Note that the fields in the outputted items are sorted the same way as they are specified in the fields parameter.
                You can use this feature to effectively fix the output format.
            omit (list of str, optional): A list of fields which should be omitted from the items.
            unwind (str, optional): Name of a field which should be unwound.
                If the field is an array then every element of the array will become a separate record and merged with parent object.
                If the unwound field is an object then it is merged with the parent object.
                If the unwound field is missing or its value is neither an array nor an object and therefore cannot be merged with a parent object,
                then the item gets preserved as it is. Note that the unwound items ignore the desc parameter.
            skip_empty (bool, optional): If True, then empty items are skipped from the output.
                Note that if used, the results might contain less items than the limit value.
            skip_hidden (bool, optional): If True, then hidden fields are skipped from the output, i.e. fields starting with the # character.
            flatten (list of str, optional): A list of fields that should be flattened
            view (str, optional): Name of the dataset view to be used

        Returns:
            ListPage: A page of the list of dataset items according to the specified filters.
        """
        dataset = await cls.open()
        return await dataset.get_data(
            offset=offset,
            limit=limit,
            desc=desc,
            clean=clean,
            fields=fields,
            omit=omit,
            unwind=unwind,
            skip_empty=skip_empty,
            skip_hidden=skip_hidden,
            flatten=flatten,
            view=view,
        )

    async def _get_data_internal(
        self,
        *,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        clean: Optional[bool] = None,
        desc: Optional[bool] = None,
        fields: Optional[List[str]] = None,
        omit: Optional[List[str]] = None,
        unwind: Optional[str] = None,
        skip_empty: Optional[bool] = None,
        skip_hidden: Optional[bool] = None,
        flatten: Optional[List[str]] = None,
        view: Optional[str] = None,
    ) -> ListPage:
        # try {
        #     return await this.client.listItems(options);
        # } catch (e) {
        #     const error = e as Error;
        #     if (error.message.includes('Cannot create a string longer than')) {
        #         throw new Error('dataset.getData(): The response is too large for parsing. You can fix this by lowering the "limit" option.');
        #     }
        #     throw e;
        # }
        # TODO: Simulate the above error in Python and handle accordingly...
        return await self._dataset_client.list_items(
            offset=offset,
            limit=limit,
            desc=desc,
            clean=clean,
            fields=fields,
            omit=omit,
            unwind=unwind,
            skip_empty=skip_empty,
            skip_hidden=skip_hidden,
            flatten=flatten,
            view=view,
        )

    async def export_to(
        self,
        key: str,
        *,
        to_key_value_store_id: Optional[str] = None,
        to_key_value_store_name: Optional[str] = None,
        content_type: Optional[str] = None,
    ) -> None:
        """Save the entirety of the dataset's contents into one file within a key-value store.

        Args:
            key (str): The key to save the data under.
            to_key_value_store_id (str, optional): The id of the key-value store in which the result will be saved.
            to_key_value_store_name (str, optional): The name of the key-value store in which the result will be saved.
                You must specify only one of `to_key_value_store_id` and `to_key_value_store_name` arguments.
                If you omit both, it uses the default key-value store.
            content_type (str, optional): Either 'text/csv' or 'application/json'. Defaults to JSON.
        """
        key_value_store = await KeyValueStore.open(id=to_key_value_store_id, name=to_key_value_store_name)
        items: List[Dict] = []
        limit = 1000
        offset = 0
        while True:
            list_items = await self._dataset_client.list_items(limit=limit, offset=offset)
            items.extend(list_items.items)
            if list_items.total <= offset + list_items.count:
                break
            offset += list_items.count

        if len(items) == 0:
            raise ValueError('Cannot export an empty dataset')

        if content_type == 'text/csv':
            output = io.StringIO()
            writer = csv.writer(output, quoting=csv.QUOTE_MINIMAL)
            writer.writerows([items[0].keys(), *[item.values() for item in items]])
            value = output.getvalue()
            return await key_value_store.set_value(key, value, content_type)

        if content_type == 'application/json':
            return await key_value_store.set_value(key, items)

        raise ValueError(f'Unsupported content type: {content_type}')

    @classmethod
    async def export_to_json(
        cls,
        key: str,
        *,
        from_dataset_id: Optional[str] = None,
        from_dataset_name: Optional[str] = None,
        to_key_value_store_id: Optional[str] = None,
        to_key_value_store_name: Optional[str] = None,
    ) -> None:
        """Save the entirety of the dataset's contents into one JSON file within a key-value store.

        Args:
            key (str): The key to save the data under.
            from_dataset_id (str, optional): The ID of the dataset in case of calling the class method. Uses default dataset if omitted.
            from_dataset_name (str, optional): The name of the dataset in case of calling the class method. Uses default dataset if omitted.
                You must specify only one of `from_dataset_id` and `from_dataset_name` arguments.
                If you omit both, it uses the default dataset.
            to_key_value_store_id (str, optional): The id of the key-value store in which the result will be saved.
            to_key_value_store_name (str, optional): The name of the key-value store in which the result will be saved.
                You must specify only one of `to_key_value_store_id` and `to_key_value_store_name` arguments.
                If you omit both, it uses the default key-value store.
        """
        dataset = await cls.open(id=from_dataset_id, name=from_dataset_name)
        await dataset.export_to_json(key, to_key_value_store_id=to_key_value_store_id, to_key_value_store_name=to_key_value_store_name)

    async def _export_to_json_internal(
        self,
        key: str,
        *,
        from_dataset_id: Optional[str] = None,  # noqa: U100
        from_dataset_name: Optional[str] = None,  # noqa: U100
        to_key_value_store_id: Optional[str] = None,
        to_key_value_store_name: Optional[str] = None,
    ) -> None:
        await self.export_to(
            key,
            to_key_value_store_id=to_key_value_store_id,
            to_key_value_store_name=to_key_value_store_name,
            content_type='application/json',
        )

    @classmethod
    async def export_to_csv(
        cls,
        key: str,
        *,
        from_dataset_id: Optional[str] = None,
        from_dataset_name: Optional[str] = None,
        to_key_value_store_id: Optional[str] = None,
        to_key_value_store_name: Optional[str] = None,
    ) -> None:
        """Save the entirety of the dataset's contents into one CSV file within a key-value store.

        Args:
            key (str): The key to save the data under.
            from_dataset_id (str, optional): The ID of the dataset in case of calling the class method. Uses default dataset if omitted.
            from_dataset_name (str, optional): The name of the dataset in case of calling the class method. Uses default dataset if omitted.
                You must specify only one of `from_dataset_id` and `from_dataset_name` arguments.
                If you omit both, it uses the default dataset.
            to_key_value_store_id (str, optional): The id of the key-value store in which the result will be saved.
            to_key_value_store_name (str, optional): The name of the key-value store in which the result will be saved.
                You must specify only one of `to_key_value_store_id` and `to_key_value_store_name` arguments.
                If you omit both, it uses the default key-value store.
        """
        dataset = await cls.open(id=from_dataset_id, name=from_dataset_name)
        await dataset.export_to_csv(key, to_key_value_store_id=to_key_value_store_id, to_key_value_store_name=to_key_value_store_name)

    async def _export_to_csv_internal(
        self,
        key: str,
        *,
        from_dataset_id: Optional[str] = None,  # noqa: U100
        from_dataset_name: Optional[str] = None,  # noqa: U100
        to_key_value_store_id: Optional[str] = None,
        to_key_value_store_name: Optional[str] = None,
    ) -> None:
        await self.export_to(
            key,
            to_key_value_store_id=to_key_value_store_id,
            to_key_value_store_name=to_key_value_store_name,
            content_type='text/csv',
        )

    async def get_info(self) -> Optional[Dict]:
        """Get an object containing general information about the dataset.

        Returns:
            dict: Object returned by calling the GET dataset API endpoint.
        """
        return await self._dataset_client.get()

    def iterate_items(
        self,
        *,
        offset: int = 0,
        limit: Optional[int] = None,
        clean: Optional[bool] = None,
        desc: Optional[bool] = None,
        fields: Optional[List[str]] = None,
        omit: Optional[List[str]] = None,
        unwind: Optional[str] = None,
        skip_empty: Optional[bool] = None,
        skip_hidden: Optional[bool] = None,
    ) -> AsyncIterator[Dict]:
        """Iterate over the items in the dataset.

        Args:
            offset (int, optional): Number of items that should be skipped at the start. The default value is 0
            limit (int, optional): Maximum number of items to return. By default there is no limit.
            desc (bool, optional): By default, results are returned in the same order as they were stored.
                To reverse the order, set this parameter to True.
            clean (bool, optional): If True, returns only non-empty items and skips hidden fields (i.e. fields starting with the # character).
                The clean parameter is just a shortcut for skip_hidden=True and skip_empty=True parameters.
                Note that since some objects might be skipped from the output, that the result might contain less items than the limit value.
            fields (list of str, optional): A list of fields which should be picked from the items,
                only these fields will remain in the resulting record objects.
                Note that the fields in the outputted items are sorted the same way as they are specified in the fields parameter.
                You can use this feature to effectively fix the output format.
            omit (list of str, optional): A list of fields which should be omitted from the items.
            unwind (str, optional): Name of a field which should be unwound.
                If the field is an array then every element of the array will become a separate record and merged with parent object.
                If the unwound field is an object then it is merged with the parent object.
                If the unwound field is missing or its value is neither an array nor an object and therefore cannot be merged with a parent object,
                then the item gets preserved as it is. Note that the unwound items ignore the desc parameter.
            skip_empty (bool, optional): If True, then empty items are skipped from the output.
                Note that if used, the results might contain less items than the limit value.
            skip_hidden (bool, optional): If True, then hidden fields are skipped from the output, i.e. fields starting with the # character.

        Yields:
            dict: An item from the dataset
        """
        return self._dataset_client.iterate_items(
            offset=offset,
            limit=limit,
            clean=clean,
            desc=desc,
            fields=fields,
            omit=omit,
            unwind=unwind,
            skip_empty=skip_empty,
            skip_hidden=skip_hidden,
        )

    async def drop(self) -> None:
        """Remove the dataset either from the Apify cloud storage or from the local directory."""
        await self._dataset_client.delete()
        self._remove_from_cache()

    @classmethod
    async def open(
        cls,
        *,
        id: Optional[str] = None,
        name: Optional[str] = None,
        force_cloud: bool = False,
        config: Optional[Configuration] = None,
    ) -> 'Dataset':
        """Open a dataset.

        Datasets are used to store structured data where each object stored has the same attributes,
        such as online store products or real estate offers.
        The actual data is stored either on the local filesystem or in the Apify cloud.

        Args:
            id (str, optional): ID of the dataset to be opened.
                If neither `id` nor `name` are provided, the method returns the default dataset associated with the actor run.
                If the dataset with the given ID does not exist, it raises an error.
            name (str, optional): Name of the dataset to be opened.
                If neither `id` nor `name` are provided, the method returns the default dataset associated with the actor run.
                If the dataset with the given name does not exist, it is created.
            force_cloud (bool, optional): If set to True, it will open a dataset on the Apify Platform even when running the actor locally.
                Defaults to False.
            config (Configuration, optional): A `Configuration` instance, uses global configuration if omitted.

        Returns:
            Dataset: An instance of the `Dataset` class for the given ID or name.
        """
        return await super().open(id=id, name=name, force_cloud=force_cloud, config=config)



from .dataset import Dataset
from .key_value_store import KeyValueStore
from .request_queue import RequestQueue
from .storage_client_manager import StorageClientManager

__all__ = ['Dataset', 'KeyValueStore', 'RequestQueue', 'StorageClientManager']



import asyncio
import io
import json
import mimetypes
import os
import pathlib
import warnings
from datetime import datetime, timezone
from operator import itemgetter
from typing import TYPE_CHECKING, Any, AsyncIterator, Dict, List, Optional

import aioshutil

from ..._crypto import _crypto_random_object_id
from ..._utils import (
    _force_rename,
    _guess_file_extension,
    _is_file_or_bytes,
    _json_dumps,
    _maybe_parse_body,
    _raise_on_duplicate_storage,
    _raise_on_non_existing_storage,
    ignore_docs,
)
from ...consts import DEFAULT_API_PARAM_LIMIT, _StorageTypes
from ...log import logger
from ..file_storage_utils import _set_or_delete_key_value_store_record, _update_metadata
from .base_resource_client import BaseResourceClient

if TYPE_CHECKING:
    from ..memory_storage_client import MemoryStorageClient

DEFAULT_LOCAL_FILE_EXTENSION = 'bin'


@ignore_docs
class KeyValueStoreClient(BaseResourceClient):
    """Sub-client for manipulating a single key-value store."""

    _id: str
    _resource_directory: str
    _memory_storage_client: 'MemoryStorageClient'
    _name: Optional[str]
    _key_value_entries: Dict[str, Dict]
    _created_at: datetime
    _accessed_at: datetime
    _modified_at: datetime
    _file_operation_lock: asyncio.Lock

    def __init__(
        self,
        *,
        base_storage_directory: str,
        memory_storage_client: 'MemoryStorageClient',
        id: Optional[str] = None,
        name: Optional[str] = None,
    ) -> None:
        """Initialize the KeyValueStoreClient."""
        self._id = id or _crypto_random_object_id()
        self._resource_directory = os.path.join(base_storage_directory, name or self._id)
        self._memory_storage_client = memory_storage_client
        self._name = name
        self._key_value_entries = {}
        self._created_at = datetime.now(timezone.utc)
        self._accessed_at = datetime.now(timezone.utc)
        self._modified_at = datetime.now(timezone.utc)
        self._file_operation_lock = asyncio.Lock()

    async def get(self) -> Optional[Dict]:
        """Retrieve the key-value store.

        Returns:
            dict, optional: The retrieved key-value store, or None if it does not exist
        """
        found = self._find_or_create_client_by_id_or_name(memory_storage_client=self._memory_storage_client, id=self._id, name=self._name)

        if found:
            async with found._file_operation_lock:
                await found._update_timestamps(False)
                return found._to_resource_info()

        return None

    async def update(self, *, name: Optional[str] = None) -> Dict:
        """Update the key-value store with specified fields.

        Args:
            name (str, optional): The new name for key-value store

        Returns:
            dict: The updated key-value store
        """
        # Check by id
        existing_store_by_id = self._find_or_create_client_by_id_or_name(
            memory_storage_client=self._memory_storage_client, id=self._id, name=self._name)

        if existing_store_by_id is None:
            _raise_on_non_existing_storage(_StorageTypes.KEY_VALUE_STORE, self._id)

        # Skip if no changes
        if name is None:
            return existing_store_by_id._to_resource_info()

        async with existing_store_by_id._file_operation_lock:
            # Check that name is not in use already
            existing_store_by_name = next(
                (store for store in self._memory_storage_client._key_value_stores_handled if store._name and store._name.lower() == name.lower()),
                None,
            )

            if existing_store_by_name is not None:
                _raise_on_duplicate_storage(_StorageTypes.KEY_VALUE_STORE, 'name', name)

            existing_store_by_id._name = name

            previous_dir = existing_store_by_id._resource_directory

            existing_store_by_id._resource_directory = os.path.join(self._memory_storage_client._key_value_stores_directory, name)

            await _force_rename(previous_dir, existing_store_by_id._resource_directory)

            # Update timestamps
            await existing_store_by_id._update_timestamps(True)

        return existing_store_by_id._to_resource_info()

    async def delete(self) -> None:
        """Delete the key-value store."""
        store = next((store for store in self._memory_storage_client._key_value_stores_handled if store._id == self._id), None)

        if store is not None:
            async with store._file_operation_lock:
                self._memory_storage_client._key_value_stores_handled.remove(store)
                store._key_value_entries.clear()

                if os.path.exists(store._resource_directory):
                    await aioshutil.rmtree(store._resource_directory)

    async def list_keys(self, *, limit: int = DEFAULT_API_PARAM_LIMIT, exclusive_start_key: Optional[str] = None) -> Dict:
        """List the keys in the key-value store.

        Args:
            limit (int, optional): Number of keys to be returned. Maximum value is 1000
            exclusive_start_key (str, optional): All keys up to this one (including) are skipped from the result

        Returns:
            dict: The list of keys in the key-value store matching the given arguments
        """
        # Check by id
        existing_store_by_id = self._find_or_create_client_by_id_or_name(
            memory_storage_client=self._memory_storage_client, id=self._id, name=self._name)

        if existing_store_by_id is None:
            _raise_on_non_existing_storage(_StorageTypes.KEY_VALUE_STORE, self._id)

        items = []

        for record in existing_store_by_id._key_value_entries.values():
            size = len(record['value'])
            items.append({
                'key': record['key'],
                'size': size,
            })

        if len(items) == 0:
            return {
                'count': len(items),
                'limit': limit,
                'exclusiveStartKey': exclusive_start_key,
                'isTruncated': False,
                'nextExclusiveStartKey': None,
                'items': items,
            }

        # Lexically sort to emulate the API
        items = sorted(items, key=itemgetter('key'))

        truncated_items = items
        if exclusive_start_key is not None:
            key_pos = next((idx for idx, i in enumerate(items) if i['key'] == exclusive_start_key), None)
            if key_pos is not None:
                truncated_items = items[key_pos + 1:]

        limited_items = truncated_items[:limit]

        last_item_in_store = items[-1]
        last_selected_item = limited_items[-1]
        is_last_selected_item_absolutely_last = last_item_in_store == last_selected_item
        next_exclusive_start_key = None if is_last_selected_item_absolutely_last else last_selected_item['key']

        async with existing_store_by_id._file_operation_lock:
            await existing_store_by_id._update_timestamps(False)

        return {
            'count': len(items),
            'limit': limit,
            'exclusiveStartKey': exclusive_start_key,
            'isTruncated': not is_last_selected_item_absolutely_last,
            'nextExclusiveStartKey': next_exclusive_start_key,
            'items': limited_items,
        }

    async def _get_record_internal(self, key: str, as_bytes: bool = False) -> Optional[Dict]:
        # Check by id
        existing_store_by_id = self._find_or_create_client_by_id_or_name(
            memory_storage_client=self._memory_storage_client, id=self._id, name=self._name)

        if existing_store_by_id is None:
            _raise_on_non_existing_storage(_StorageTypes.KEY_VALUE_STORE, self._id)

        entry = existing_store_by_id._key_value_entries.get(key)

        if entry is None:
            return None

        record = {
            'key': entry['key'],
            'value': entry['value'],
            # To guess the type, we need a real file name, not just the extension. e.g. 'file.json' instead of 'json'
            'contentType': entry.get('content_type') or mimetypes.guess_type(f"file.{entry['extension']}")[0],
        }

        if not as_bytes:
            try:
                record['value'] = _maybe_parse_body(record['value'], record['contentType'])
            except ValueError:
                logger.exception('Error parsing key-value store record')

        async with existing_store_by_id._file_operation_lock:
            await existing_store_by_id._update_timestamps(False)

        return record

    async def get_record(self, key: str) -> Optional[Dict]:
        """Retrieve the given record from the key-value store.

        Args:
            key (str): Key of the record to retrieve

        Returns:
            dict, optional: The requested record, or None, if the record does not exist
        """
        return await self._get_record_internal(key)

    async def get_record_as_bytes(self, key: str) -> Optional[Dict]:
        """Retrieve the given record from the key-value store, without parsing it.

        Args:
            key (str): Key of the record to retrieve

        Returns:
            dict, optional: The requested record, or None, if the record does not exist
        """
        return await self._get_record_internal(key, as_bytes=True)

    async def stream_record(self, _key: str) -> AsyncIterator[Optional[Dict]]:  # noqa: D102
        raise NotImplementedError('This method is not supported in local memory storage.')

    async def set_record(self, key: str, value: Any, content_type: Optional[str] = None) -> None:
        """Set a value to the given record in the key-value store.

        Args:
            key (str): The key of the record to save the value to
            value (Any): The value to save into the record
            content_type (str, optional): The content type of the saved value
        """
        # Check by id
        existing_store_by_id = self._find_or_create_client_by_id_or_name(
            memory_storage_client=self._memory_storage_client, id=self._id, name=self._name)

        if existing_store_by_id is None:
            _raise_on_non_existing_storage(_StorageTypes.KEY_VALUE_STORE, self._id)

        if isinstance(value, io.IOBase):
            raise NotImplementedError('File-like values are not supported in local memory storage')

        if content_type is None:
            if _is_file_or_bytes(value):
                content_type = 'application/octet-stream'
            elif isinstance(value, str):
                content_type = 'text/plain; charset=utf-8'
            else:
                content_type = 'application/json; charset=utf-8'

        extension = _guess_file_extension(content_type or '') or DEFAULT_LOCAL_FILE_EXTENSION

        if 'application/json' in content_type and not _is_file_or_bytes(value) and not isinstance(value, str):
            value = _json_dumps(value).encode('utf-8')

        async with existing_store_by_id._file_operation_lock:
            record = {
                'extension': extension,
                'key': key,
                'value': value,
                'content_type': content_type,
            }

            existing_store_by_id._key_value_entries[key] = record

            await existing_store_by_id._update_timestamps(True)
            await _set_or_delete_key_value_store_record(
                entity_directory=existing_store_by_id._resource_directory,
                persist_storage=self._memory_storage_client._persist_storage,
                record=record,
                should_set=True,
                write_metadata=self._memory_storage_client._write_metadata,
            )

    async def delete_record(self, key: str) -> None:
        """Delete the specified record from the key-value store.

        Args:
            key (str): The key of the record which to delete
        """
        # Check by id
        existing_store_by_id = self._find_or_create_client_by_id_or_name(
            memory_storage_client=self._memory_storage_client, id=self._id, name=self._name)

        if existing_store_by_id is None:
            _raise_on_non_existing_storage(_StorageTypes.KEY_VALUE_STORE, self._id)

        entry = existing_store_by_id._key_value_entries.get(key)

        if entry is not None:
            async with existing_store_by_id._file_operation_lock:
                del existing_store_by_id._key_value_entries[key]
                await existing_store_by_id._update_timestamps(True)
                await _set_or_delete_key_value_store_record(
                    entity_directory=existing_store_by_id._resource_directory,
                    persist_storage=self._memory_storage_client._persist_storage,
                    record=entry,
                    should_set=False,
                    write_metadata=self._memory_storage_client._write_metadata,
                )

    def _to_resource_info(self) -> Dict:
        """Retrieve the key-value store info."""
        return {
            'id': self._id,
            'name': self._name,
            'accessedAt': self._accessed_at,
            'createdAt': self._created_at,
            'modifiedAt': self._modified_at,
            'userId': '1',
        }

    async def _update_timestamps(self, has_been_modified: bool) -> None:
        self._accessed_at = datetime.now(timezone.utc)

        if has_been_modified:
            self._modified_at = datetime.now(timezone.utc)

        kv_store_info = self._to_resource_info()
        await _update_metadata(
            data=kv_store_info,
            entity_directory=self._resource_directory,
            write_metadata=self._memory_storage_client._write_metadata,
        )

    @classmethod
    def _get_storages_dir(cls, memory_storage_client: 'MemoryStorageClient') -> str:
        return memory_storage_client._key_value_stores_directory

    @classmethod
    def _get_storage_client_cache(cls, memory_storage_client: 'MemoryStorageClient') -> List['KeyValueStoreClient']:
        return memory_storage_client._key_value_stores_handled

    @classmethod
    def _create_from_directory(
        cls,
        storage_directory: str,
        memory_storage_client: 'MemoryStorageClient',
        id: Optional[str] = None,
        name: Optional[str] = None,
    ) -> 'KeyValueStoreClient':
        created_at = datetime.now(timezone.utc)
        accessed_at = datetime.now(timezone.utc)
        modified_at = datetime.now(timezone.utc)
        internal_records: Dict[str, Dict] = {}

        # Access the key value store folder
        for entry in os.scandir(storage_directory):
            if entry.is_file():
                if entry.name == '__metadata__.json':
                    # We have found the store metadata file, build out information based on it
                    with open(os.path.join(storage_directory, entry.name), encoding='utf8') as f:
                        metadata = json.load(f)
                    id = metadata['id']
                    name = metadata['name']
                    created_at = datetime.fromisoformat(metadata['createdAt'])
                    accessed_at = datetime.fromisoformat(metadata['accessedAt'])
                    modified_at = datetime.fromisoformat(metadata['modifiedAt'])

                    continue

                if '.__metadata__.' in entry.name:
                    # This is an entry's metadata file, we can use it to create/extend the record
                    with open(os.path.join(storage_directory, entry.name), encoding='utf8') as f:
                        metadata = json.load(f)

                    new_record = {
                        **internal_records.get(metadata['key'], {}),
                        **metadata,
                    }

                    internal_records[metadata['key']] = new_record

                    continue

                with open(os.path.join(storage_directory, entry.name), 'rb') as f:
                    file_content = f.read()
                file_extension = pathlib.Path(entry.name).suffix
                content_type, _ = mimetypes.guess_type(entry.name)
                if content_type is None:
                    content_type = 'text/plain'
                extension = _guess_file_extension(content_type)

                if file_extension == '':
                    # We need to override and then restore the warnings filter so that the warning gets printed out,
                    # Otherwise it would be silently swallowed
                    with warnings.catch_warnings():
                        warnings.simplefilter('always')
                        warnings.warn(
                            f"""Key-value entry "{entry.name}" for store {name or id} does not have a file extension, assuming it as text.
                            If you want to have correct interpretation of the file, you should add a file extension to the entry.""",
                            Warning,
                            stacklevel=2,
                        )
                elif 'application/json' in content_type:
                    try:
                        # Try parsing the JSON ahead of time (not ideal but solves invalid files being loaded into stores)
                        json.loads(file_content)
                    except json.JSONDecodeError:
                        # We need to override and then restore the warnings filter so that the warning gets printed out,
                        # Otherwise it would be silently swallowed
                        with warnings.catch_warnings():
                            warnings.simplefilter('always')
                            warnings.warn(
                                (f'Key-value entry "{entry.name}" for store {name or id} has invalid JSON content'
                                    'and will be ignored from the store.'),
                                Warning,
                                stacklevel=2,
                            )
                        continue

                name_split = entry.name.split('.')

                if file_extension != '':
                    name_split.pop()

                key = '.'.join(name_split)

                new_record = {
                    'key': key,
                    'extension': extension,
                    'value': file_content,
                    'content_type': content_type,
                    **internal_records.get(key, {}),
                }

                internal_records[key] = new_record

        new_client = KeyValueStoreClient(
            base_storage_directory=memory_storage_client._key_value_stores_directory,
            memory_storage_client=memory_storage_client,
            id=id,
            name=name,
        )

        # Overwrite properties
        new_client._accessed_at = accessed_at
        new_client._created_at = created_at
        new_client._modified_at = modified_at

        for key, record in internal_records.items():
            new_client._key_value_entries[key] = record

        return new_client



from typing import Optional, Union

from apify_client import ApifyClientAsync

from .._memory_storage import MemoryStorageClient
from .._utils import ignore_docs
from ..config import Configuration


@ignore_docs
class StorageClientManager:
    """A class for managing storage clients."""

    _config: Configuration

    _local_client: Optional[MemoryStorageClient] = None
    _cloud_client: Optional[ApifyClientAsync] = None

    _default_instance: Optional['StorageClientManager'] = None

    def __init__(self) -> None:
        """Create a `StorageClientManager` instance."""
        self._config = Configuration.get_global_configuration()

    @classmethod
    def set_config(cls, config: Configuration) -> None:
        """Set the config for the StorageClientManager.

        Args:
            config (Configuration): The configuration this StorageClientManager should use.
        """
        cls._get_default_instance()._config = config

    @classmethod
    def get_storage_client(cls, force_cloud: bool = False) -> Union[ApifyClientAsync, MemoryStorageClient]:
        """Get the current storage client instance.

        Returns:
            ApifyClientAsync or MemoryStorageClient: The current storage client instance.
        """
        default_instance = cls._get_default_instance()
        if not default_instance._local_client:
            default_instance._local_client = MemoryStorageClient(persist_storage=default_instance._config.persist_storage, write_metadata=True)

        if default_instance._config.is_at_home or force_cloud:
            assert default_instance._cloud_client is not None
            return default_instance._cloud_client

        return default_instance._local_client

    @classmethod
    def set_cloud_client(cls, client: ApifyClientAsync) -> None:
        """Set the storage client.

        Args:
            client (ApifyClientAsync or MemoryStorageClient): The instance of a storage client.
        """
        cls._get_default_instance()._cloud_client = client

    @classmethod
    def _get_default_instance(cls) -> 'StorageClientManager':
        if cls._default_instance is None:
            cls._default_instance = cls()

        return cls._default_instance



import asyncio
import json
import os
from datetime import datetime, timezone
from decimal import Decimal
from typing import TYPE_CHECKING, Dict, List, Optional

import aioshutil
from sortedcollections import ValueSortedDict  # type: ignore

from ..._crypto import _crypto_random_object_id
from ..._utils import (
    _filter_out_none_values_recursively,
    _force_rename,
    _json_dumps,
    _raise_on_duplicate_storage,
    _raise_on_non_existing_storage,
    _unique_key_to_request_id,
    ignore_docs,
)
from ...consts import _StorageTypes
from ..file_storage_utils import _delete_request, _update_metadata, _update_request_queue_item
from .base_resource_client import BaseResourceClient

if TYPE_CHECKING:
    from ..memory_storage_client import MemoryStorageClient


@ignore_docs
class RequestQueueClient(BaseResourceClient):
    """Sub-client for manipulating a single request queue."""

    _id: str
    _resource_directory: str
    _memory_storage_client: 'MemoryStorageClient'
    _name: Optional[str]
    _requests: ValueSortedDict
    _created_at: datetime
    _accessed_at: datetime
    _modified_at: datetime
    _handled_request_count = 0
    _pending_request_count = 0
    _last_used_timestamp = Decimal(0.0)
    _file_operation_lock: asyncio.Lock

    def __init__(
        self,
        *,
        base_storage_directory: str,
        memory_storage_client: 'MemoryStorageClient',
        id: Optional[str] = None,
        name: Optional[str] = None,
    ) -> None:
        """Initialize the RequestQueueClient."""
        self._id = id or _crypto_random_object_id()
        self._resource_directory = os.path.join(base_storage_directory, name or self._id)
        self._memory_storage_client = memory_storage_client
        self._name = name
        self._requests = ValueSortedDict(lambda req: req.get('orderNo') or -float('inf'))
        self._created_at = datetime.now(timezone.utc)
        self._accessed_at = datetime.now(timezone.utc)
        self._modified_at = datetime.now(timezone.utc)
        self._file_operation_lock = asyncio.Lock()

    async def get(self) -> Optional[Dict]:
        """Retrieve the request queue.

        Returns:
            dict, optional: The retrieved request queue, or None, if it does not exist
        """
        found = self._find_or_create_client_by_id_or_name(memory_storage_client=self._memory_storage_client, id=self._id, name=self._name)

        if found:
            async with found._file_operation_lock:
                await found._update_timestamps(False)
                return found._to_resource_info()

        return None

    async def update(self, *, name: Optional[str] = None) -> Dict:
        """Update the request queue with specified fields.

        Args:
            name (str, optional): The new name for the request queue

        Returns:
            dict: The updated request queue
        """
        # Check by id
        existing_queue_by_id = self._find_or_create_client_by_id_or_name(
            memory_storage_client=self._memory_storage_client, id=self._id, name=self._name)

        if existing_queue_by_id is None:
            _raise_on_non_existing_storage(_StorageTypes.REQUEST_QUEUE, self._id)

        # Skip if no changes
        if name is None:
            return existing_queue_by_id._to_resource_info()

        async with existing_queue_by_id._file_operation_lock:
            # Check that name is not in use already
            existing_queue_by_name = next(
                (queue for queue in self._memory_storage_client._request_queues_handled if queue._name and queue._name.lower() == name.lower()), None)

            if existing_queue_by_name is not None:
                _raise_on_duplicate_storage(_StorageTypes.REQUEST_QUEUE, 'name', name)

            existing_queue_by_id._name = name

            previous_dir = existing_queue_by_id._resource_directory

            existing_queue_by_id._resource_directory = os.path.join(self._memory_storage_client._request_queues_directory, name)

            await _force_rename(previous_dir, existing_queue_by_id._resource_directory)

            # Update timestamps
            await existing_queue_by_id._update_timestamps(True)

            return existing_queue_by_id._to_resource_info()

    async def delete(self) -> None:
        """Delete the request queue."""
        queue = next((queue for queue in self._memory_storage_client._request_queues_handled if queue._id == self._id), None)

        if queue is not None:
            async with queue._file_operation_lock:
                self._memory_storage_client._request_queues_handled.remove(queue)
                queue._pending_request_count = 0
                queue._handled_request_count = 0
                queue._requests.clear()

                if os.path.exists(queue._resource_directory):
                    await aioshutil.rmtree(queue._resource_directory)

    async def list_head(self, *, limit: Optional[int] = None) -> Dict:
        """Retrieve a given number of requests from the beginning of the queue.

        Args:
            limit (int, optional): How many requests to retrieve

        Returns:
            dict: The desired number of requests from the beginning of the queue.
        """
        existing_queue_by_id = self._find_or_create_client_by_id_or_name(
            memory_storage_client=self._memory_storage_client, id=self._id, name=self._name)

        if existing_queue_by_id is None:
            _raise_on_non_existing_storage(_StorageTypes.REQUEST_QUEUE, self._id)

        async with existing_queue_by_id._file_operation_lock:
            await existing_queue_by_id._update_timestamps(False)

            items: List[Dict] = []

            # Iterate all requests in the queue which have sorted key larger than infinity, which means `orderNo` is not `None`
            # This will iterate them in order of `orderNo`
            for request_key in existing_queue_by_id._requests.irange_key(min_key=-float('inf'), inclusive=(False, True)):
                if len(items) == limit:
                    break

                request = existing_queue_by_id._requests.get(request_key)

                # Check that the request still exists and was not handled,
                # in case something deleted it or marked it as handled concurrenctly
                if request and request['orderNo']:
                    items.append(request)

            return {
                'limit': limit,
                'hadMultipleClients': False,
                'queueModifiedAt': existing_queue_by_id._modified_at,
                'items': [self._json_to_request(item['json']) for item in items],
            }

    async def add_request(self, request: Dict, *, forefront: Optional[bool] = None) -> Dict:
        """Add a request to the queue.

        Args:
            request (dict): The request to add to the queue
            forefront (bool, optional): Whether to add the request to the head or the end of the queue

        Returns:
            dict: The added request.
        """
        existing_queue_by_id = self._find_or_create_client_by_id_or_name(
            memory_storage_client=self._memory_storage_client, id=self._id, name=self._name)

        if existing_queue_by_id is None:
            _raise_on_non_existing_storage(_StorageTypes.REQUEST_QUEUE, self._id)

        request_model = self._create_internal_request(request, forefront)

        async with existing_queue_by_id._file_operation_lock:
            existing_request_with_id = existing_queue_by_id._requests.get(request_model['id'])

            # We already have the request present, so we return information about it
            if existing_request_with_id is not None:
                await existing_queue_by_id._update_timestamps(False)

                return {
                    'requestId': existing_request_with_id['id'],
                    'wasAlreadyHandled': existing_request_with_id['orderNo'] is None,
                    'wasAlreadyPresent': True,
                }

            existing_queue_by_id._requests[request_model['id']] = request_model
            if request_model['orderNo'] is None:
                existing_queue_by_id._handled_request_count += 1
            else:
                existing_queue_by_id._pending_request_count += 1
            await existing_queue_by_id._update_timestamps(True)
            await _update_request_queue_item(
                request=request_model,
                request_id=request_model['id'],
                entity_directory=existing_queue_by_id._resource_directory,
                persist_storage=self._memory_storage_client._persist_storage,
            )

            return {
                'requestId': request_model['id'],
                # We return wasAlreadyHandled: false even though the request may
                # have been added as handled, because that's how API behaves.
                'wasAlreadyHandled': False,
                'wasAlreadyPresent': False,
            }

    async def get_request(self, request_id: str) -> Optional[Dict]:
        """Retrieve a request from the queue.

        Args:
            request_id (str): ID of the request to retrieve

        Returns:
            dict, optional: The retrieved request, or None, if it did not exist.
        """
        existing_queue_by_id = self._find_or_create_client_by_id_or_name(
            memory_storage_client=self._memory_storage_client, id=self._id, name=self._name)

        if existing_queue_by_id is None:
            _raise_on_non_existing_storage(_StorageTypes.REQUEST_QUEUE, self._id)

        async with existing_queue_by_id._file_operation_lock:
            await existing_queue_by_id._update_timestamps(False)

            request = existing_queue_by_id._requests.get(request_id)
            return self._json_to_request(request['json'] if request is not None else None)

    async def update_request(self, request: Dict, *, forefront: Optional[bool] = None) -> Dict:
        """Update a request in the queue.

        Args:
            request (dict): The updated request
            forefront (bool, optional): Whether to put the updated request in the beginning or the end of the queue

        Returns:
            dict: The updated request
        """
        existing_queue_by_id = self._find_or_create_client_by_id_or_name(
            memory_storage_client=self._memory_storage_client, id=self._id, name=self._name)

        if existing_queue_by_id is None:
            _raise_on_non_existing_storage(_StorageTypes.REQUEST_QUEUE, self._id)

        request_model = self._create_internal_request(request, forefront)

        # First we need to check the existing request to be
        # able to return information about its handled state.

        existing_request = existing_queue_by_id._requests.get(request_model['id'])

        # Undefined means that the request is not present in the queue.
        # We need to insert it, to behave the same as API.
        if existing_request is None:
            return await self.add_request(request, forefront=forefront)

        async with existing_queue_by_id._file_operation_lock:
            # When updating the request, we need to make sure that
            # the handled counts are updated correctly in all cases.
            existing_queue_by_id._requests[request_model['id']] = request_model

            pending_count_adjustment = 0
            is_request_handled_state_changing = type(existing_request['orderNo']) != type(request_model['orderNo'])  # noqa
            request_was_handled_before_update = existing_request['orderNo'] is None

            # We add 1 pending request if previous state was handled
            if is_request_handled_state_changing:
                pending_count_adjustment = 1 if request_was_handled_before_update else -1

            existing_queue_by_id._pending_request_count += pending_count_adjustment
            existing_queue_by_id._handled_request_count -= pending_count_adjustment
            await existing_queue_by_id._update_timestamps(True)
            await _update_request_queue_item(
                request=request_model,
                request_id=request_model['id'],
                entity_directory=existing_queue_by_id._resource_directory,
                persist_storage=self._memory_storage_client._persist_storage,
            )

            return {
                'requestId': request_model['id'],
                'wasAlreadyHandled': request_was_handled_before_update,
                'wasAlreadyPresent': True,
            }

    async def delete_request(self, request_id: str) -> None:
        """Delete a request from the queue.

        Args:
            request_id (str): ID of the request to delete.
        """
        existing_queue_by_id = self._find_or_create_client_by_id_or_name(
            memory_storage_client=self._memory_storage_client, id=self._id, name=self._name)

        if existing_queue_by_id is None:
            _raise_on_non_existing_storage(_StorageTypes.REQUEST_QUEUE, self._id)

        async with existing_queue_by_id._file_operation_lock:
            request = existing_queue_by_id._requests.get(request_id)

            if request:
                del existing_queue_by_id._requests[request_id]
                if request['orderNo'] is None:
                    existing_queue_by_id._handled_request_count -= 1
                else:
                    existing_queue_by_id._pending_request_count -= 1
                await existing_queue_by_id._update_timestamps(True)
                await _delete_request(entity_directory=existing_queue_by_id._resource_directory, request_id=request_id)

    def _to_resource_info(self) -> Dict:
        """Retrieve the request queue store info."""
        return {
            'accessedAt': self._accessed_at,
            'createdAt': self._created_at,
            'hadMultipleClients': False,
            'handledRequestCount': self._handled_request_count,
            'id': self._id,
            'modifiedAt': self._modified_at,
            'name': self._name,
            'pendingRequestCount': self._pending_request_count,
            'stats': {},
            'totalRequestCount': len(self._requests),
            'userId': '1',
        }

    async def _update_timestamps(self, has_been_modified: bool) -> None:
        self._accessed_at = datetime.now(timezone.utc)

        if has_been_modified:
            self._modified_at = datetime.now(timezone.utc)

        request_queue_info = self._to_resource_info()
        await _update_metadata(
            data=request_queue_info,
            entity_directory=self._resource_directory,
            write_metadata=self._memory_storage_client._write_metadata,
        )

    def _json_to_request(self, request_json: Optional[str]) -> Optional[dict]:
        if request_json is None:
            return None
        request = json.loads(request_json)
        return _filter_out_none_values_recursively(request)

    def _create_internal_request(self, request: Dict, forefront: Optional[bool]) -> Dict:
        order_no = self._calculate_order_no(request, forefront)
        id = _unique_key_to_request_id(request['uniqueKey'])

        if request.get('id') is not None and request['id'] != id:
            raise ValueError('Request ID does not match its unique_key.')

        json_request = _json_dumps({**request, 'id': id})
        return {
            'id': id,
            'json': json_request,
            'method': request.get('method'),
            'orderNo': order_no,
            'retryCount': request.get('retryCount', 0),
            'uniqueKey': request['uniqueKey'],
            'url': request['url'],
        }

    def _calculate_order_no(self, request: Dict, forefront: Optional[bool]) -> Optional[Decimal]:
        if request.get('handledAt') is not None:
            return None

        # Get the current timestamp in milliseconds
        timestamp = Decimal(datetime.now(timezone.utc).timestamp()) * 1000
        timestamp = round(timestamp, 6)

        # Make sure that this timestamp was not used yet, so that we have unique orderNos
        if timestamp <= self._last_used_timestamp:
            timestamp = self._last_used_timestamp + Decimal(0.000001)

        self._last_used_timestamp = timestamp

        return -timestamp if forefront else timestamp

    @classmethod
    def _get_storages_dir(cls, memory_storage_client: 'MemoryStorageClient') -> str:
        return memory_storage_client._request_queues_directory

    @classmethod
    def _get_storage_client_cache(cls, memory_storage_client: 'MemoryStorageClient') -> List['RequestQueueClient']:
        return memory_storage_client._request_queues_handled

    @classmethod
    def _create_from_directory(
        cls,
        storage_directory: str,
        memory_storage_client: 'MemoryStorageClient',
        id: Optional[str] = None,
        name: Optional[str] = None,
    ) -> 'RequestQueueClient':
        created_at = datetime.now(timezone.utc)
        accessed_at = datetime.now(timezone.utc)
        modified_at = datetime.now(timezone.utc)
        handled_request_count = 0
        pending_request_count = 0
        entries: List[Dict] = []

        # Access the request queue folder
        for entry in os.scandir(storage_directory):
            if entry.is_file():
                if entry.name == '__metadata__.json':
                    # We have found the queue's metadata file, build out information based on it
                    with open(os.path.join(storage_directory, entry.name)) as f:
                        metadata = json.load(f)
                    id = metadata['id']
                    name = metadata['name']
                    created_at = datetime.fromisoformat(metadata['createdAt'])
                    accessed_at = datetime.fromisoformat(metadata['accessedAt'])
                    modified_at = datetime.fromisoformat(metadata['modifiedAt'])
                    handled_request_count = metadata['handledRequestCount']
                    pending_request_count = metadata['pendingRequestCount']

                    continue

                with open(os.path.join(storage_directory, entry.name)) as f:
                    request = json.load(f)
                entries.append(request)

        new_client = cls(
            base_storage_directory=memory_storage_client._request_queues_directory,
            memory_storage_client=memory_storage_client,
            id=id,
            name=name,
        )

        # Overwrite properties
        new_client._accessed_at = accessed_at
        new_client._created_at = created_at
        new_client._modified_at = modified_at
        new_client._handled_request_count = handled_request_count
        new_client._pending_request_count = pending_request_count

        for request in entries:
            new_client._requests[request['id']] = request

        return new_client



import json
import os
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Dict, List, Optional

from typing_extensions import Self

from ..._utils import ignore_docs

if TYPE_CHECKING:
    from ..memory_storage_client import MemoryStorageClient


@ignore_docs
class BaseResourceClient(ABC):
    """Base class for resource clients."""

    _id: str
    _name: Optional[str]
    _resource_directory: str

    @abstractmethod
    def __init__(
        self,
        *,
        base_storage_directory: str,
        memory_storage_client: 'MemoryStorageClient',
        id: Optional[str] = None,
        name: Optional[str] = None,
    ) -> None:
        """Initialize the BaseResourceClient."""
        raise NotImplementedError('You must override this method in the subclass!')

    @abstractmethod
    async def get(self) -> Optional[Dict]:
        """Retrieve the storage.

        Returns:
            dict, optional: The retrieved storage, or None, if it does not exist
        """
        raise NotImplementedError('You must override this method in the subclass!')

    @classmethod
    @abstractmethod
    def _get_storages_dir(cls, memory_storage_client: 'MemoryStorageClient') -> str:
        raise NotImplementedError('You must override this method in the subclass!')

    @classmethod
    @abstractmethod
    def _get_storage_client_cache(cls, memory_storage_client: 'MemoryStorageClient') -> List[Self]:
        raise NotImplementedError('You must override this method in the subclass!')

    @abstractmethod
    def _to_resource_info(self) -> Dict:
        raise NotImplementedError('You must override this method in the subclass!')

    @classmethod
    @abstractmethod
    def _create_from_directory(
        cls,
        storage_directory: str,
        memory_storage_client: 'MemoryStorageClient',
        id: Optional[str] = None,
        name: Optional[str] = None,
    ) -> Self:
        raise NotImplementedError('You must override this method in the subclass!')

    @classmethod
    def _find_or_create_client_by_id_or_name(
        cls,
        memory_storage_client: 'MemoryStorageClient',
        id: Optional[str] = None,
        name: Optional[str] = None,
    ) -> Optional[Self]:
        assert id is not None or name is not None

        storage_client_cache = cls._get_storage_client_cache(memory_storage_client)
        storages_dir = cls._get_storages_dir(memory_storage_client)

        # First check memory cache
        found = next((storage_client for storage_client in storage_client_cache
                      if storage_client._id == id or (storage_client._name and name and storage_client._name.lower() == name.lower())), None)

        if found is not None:
            return found

        storage_path = None

        # First try to find the storage by looking up the directory by name
        if name:
            possible_storage_path = os.path.join(storages_dir, name)
            if os.access(possible_storage_path, os.F_OK):
                storage_path = possible_storage_path

        # If it's not found, try going through the storages dir and finding it by metadata
        if not storage_path:
            if os.access(storages_dir, os.F_OK):
                for entry in os.scandir(storages_dir):
                    if not entry.is_dir():
                        continue
                    metadata_path = os.path.join(entry.path, '__metadata__.json')
                    if not os.access(metadata_path, os.F_OK):
                        continue
                    with open(metadata_path) as metadata_file:
                        metadata = json.load(metadata_file)
                    if id and id == metadata.get('id'):
                        storage_path = entry.path
                        name = metadata.get(name)
                        break
                    if name and name == metadata.get('name'):
                        storage_path = entry.path
                        id = metadata.get(id)
                        break

        # As a last resort, try to check if the accessed storage is the default one,
        # and the folder has no metadata
        # TODO: make this respect the APIFY_DEFAULT_XXX_ID env var
        if id == 'default':
            possible_storage_path = os.path.join(storages_dir, id)
            if os.access(possible_storage_path, os.F_OK):
                storage_path = possible_storage_path

        if not storage_path:
            return None

        resource_client = cls._create_from_directory(storage_path, memory_storage_client, id, name)

        storage_client_cache.append(resource_client)

        return resource_client



import base64
import secrets
from typing import Any

from cryptography.exceptions import InvalidTag as InvalidTagException
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from ._utils import ignore_docs
from .consts import ENCRYPTED_INPUT_VALUE_REGEXP

ENCRYPTION_KEY_LENGTH = 32
ENCRYPTION_IV_LENGTH = 16
ENCRYPTION_AUTH_TAG_LENGTH = 16


@ignore_docs
def public_encrypt(value: str, *, public_key: rsa.RSAPublicKey) -> dict:
    """Encrypts the given value using AES cipher and the password for encryption using the public key.

    The encryption password is a string of encryption key and initial vector used for cipher.
    It returns the encrypted password and encrypted value in BASE64 format.

    Args:
        value (str): The value which should be encrypted.
        public_key (RSAPublicKey): Public key to use for encryption.

    Returns:
        disc: Encrypted password and value.
    """
    key_bytes = _crypto_random_object_id(ENCRYPTION_KEY_LENGTH).encode('utf-8')
    initialized_vector_bytes = _crypto_random_object_id(ENCRYPTION_IV_LENGTH).encode('utf-8')
    value_bytes = value.encode('utf-8')

    password_bytes = key_bytes + initialized_vector_bytes

    # NOTE: Auth Tag is appended to the end of the encrypted data, it has length of 16 bytes and ensures integrity of the data.
    cipher = Cipher(algorithms.AES(key_bytes), modes.GCM(initialized_vector_bytes, min_tag_length=ENCRYPTION_AUTH_TAG_LENGTH))
    encryptor = cipher.encryptor()
    encrypted_value_bytes = encryptor.update(value_bytes) + encryptor.finalize()
    encrypted_password_bytes = public_key.encrypt(
        password_bytes,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA1()),
            algorithm=hashes.SHA1(),
            label=None,
        ),
    )
    return {
        'encrypted_value': base64.b64encode(encrypted_value_bytes + encryptor.tag).decode('utf-8'),
        'encrypted_password': base64.b64encode(encrypted_password_bytes).decode('utf-8'),
    }


@ignore_docs
def private_decrypt(
    encrypted_password: str,
    encrypted_value: str,
    *,
    private_key: rsa.RSAPrivateKey,
) -> str:
    """Decrypts the given encrypted value using the private key and password.

    Args:
        encrypted_password (str): Password used to encrypt the private key encoded as base64 string.
        encrypted_value (str): Encrypted value to decrypt as base64 string.
        private_key (RSAPrivateKey): Private key to use for decryption.

    Returns:
        str: Decrypted value.
    """
    encrypted_password_bytes = base64.b64decode(encrypted_password.encode('utf-8'))
    encrypted_value_bytes = base64.b64decode(encrypted_value.encode('utf-8'))

    # Decrypt the password
    password_bytes = private_key.decrypt(
        encrypted_password_bytes,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA1()),
            algorithm=hashes.SHA1(),
            label=None,
        ),
    )

    if len(password_bytes) != ENCRYPTION_KEY_LENGTH + ENCRYPTION_IV_LENGTH:
        raise ValueError('Decryption failed, invalid password length!')

    # Slice the encrypted into cypher and authentication tag
    authentication_tag_bytes = encrypted_value_bytes[-ENCRYPTION_AUTH_TAG_LENGTH:]
    encrypted_data_bytes = encrypted_value_bytes[:len(encrypted_value_bytes) - ENCRYPTION_AUTH_TAG_LENGTH]
    encryption_key_bytes = password_bytes[:ENCRYPTION_KEY_LENGTH]
    initialization_vector_bytes = password_bytes[ENCRYPTION_KEY_LENGTH:]

    try:
        cipher = Cipher(algorithms.AES(encryption_key_bytes), modes.GCM(initialization_vector_bytes, authentication_tag_bytes))
        decryptor = cipher.decryptor()
        decipher_bytes = decryptor.update(encrypted_data_bytes) + decryptor.finalize()
    except InvalidTagException:
        raise ValueError('Decryption failed, malformed encrypted value or password.')
    except Exception as err:
        raise err

    return decipher_bytes.decode('utf-8')


def _load_private_key(private_key_file_base64: str, private_key_password: str) -> rsa.RSAPrivateKey:
    private_key = serialization.load_pem_private_key(base64.b64decode(
        private_key_file_base64.encode('utf-8')), password=private_key_password.encode('utf-8'))
    if not isinstance(private_key, rsa.RSAPrivateKey):
        raise ValueError('Invalid private key.')

    return private_key


def _load_public_key(public_key_file_base64: str) -> rsa.RSAPublicKey:
    public_key = serialization.load_pem_public_key(base64.b64decode(public_key_file_base64.encode('utf-8')))
    if not isinstance(public_key, rsa.RSAPublicKey):
        raise ValueError('Invalid public key.')

    return public_key


def _crypto_random_object_id(length: int = 17) -> str:
    """Python reimplementation of cryptoRandomObjectId from `@apify/utilities`."""
    chars = 'abcdefghijklmnopqrstuvwxyzABCEDFGHIJKLMNOPQRSTUVWXYZ0123456789'
    return ''.join(secrets.choice(chars) for _ in range(length))


def _decrypt_input_secrets(private_key: rsa.RSAPrivateKey, input: Any) -> Any:
    """Decrypt input secrets."""
    if not isinstance(input, dict):
        return input

    for key, value in input.items():
        if isinstance(value, str):
            match = ENCRYPTED_INPUT_VALUE_REGEXP.fullmatch(value)
            if match:
                encrypted_password = match.group(1)
                encrypted_value = match.group(2)
                input[key] = private_decrypt(
                    encrypted_password,
                    encrypted_value,
                    private_key=private_key,
                )

    return input



import asyncio
import json
import os
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any, AsyncIterator, Dict, List, Optional, Tuple

import aioshutil

from apify_client._utils import ListPage

from ..._crypto import _crypto_random_object_id
from ..._types import JSONSerializable
from ..._utils import _force_rename, _raise_on_duplicate_storage, _raise_on_non_existing_storage, ignore_docs
from ...consts import _StorageTypes
from ..file_storage_utils import _update_dataset_items, _update_metadata
from .base_resource_client import BaseResourceClient

if TYPE_CHECKING:
    from ..memory_storage_client import MemoryStorageClient

# This is what API returns in the x-apify-pagination-limit
# header when no limit query parameter is used.
LIST_ITEMS_LIMIT = 999_999_999_999

# Number of characters of the dataset item file names.
# E.g.: 000000019.json - 9 digits
LOCAL_ENTRY_NAME_DIGITS = 9


@ignore_docs
class DatasetClient(BaseResourceClient):
    """Sub-client for manipulating a single dataset."""

    _id: str
    _resource_directory: str
    _memory_storage_client: 'MemoryStorageClient'
    _name: Optional[str]
    _dataset_entries: Dict[str, Dict]
    _created_at: datetime
    _accessed_at: datetime
    _modified_at: datetime
    _item_count = 0
    _file_operation_lock: asyncio.Lock

    def __init__(
        self,
        *,
        base_storage_directory: str,
        memory_storage_client: 'MemoryStorageClient',
        id: Optional[str] = None,
        name: Optional[str] = None,
    ) -> None:
        """Initialize the DatasetClient."""
        self._id = id or _crypto_random_object_id()
        self._resource_directory = os.path.join(base_storage_directory, name or self._id)
        self._memory_storage_client = memory_storage_client
        self._name = name
        self._dataset_entries = {}
        self._created_at = datetime.now(timezone.utc)
        self._accessed_at = datetime.now(timezone.utc)
        self._modified_at = datetime.now(timezone.utc)
        self._file_operation_lock = asyncio.Lock()

    async def get(self) -> Optional[Dict]:
        """Retrieve the dataset.

        Returns:
            dict, optional: The retrieved dataset, or None, if it does not exist
        """
        found = self._find_or_create_client_by_id_or_name(memory_storage_client=self._memory_storage_client, id=self._id, name=self._name)

        if found:
            async with found._file_operation_lock:
                await found._update_timestamps(False)
                return found._to_resource_info()

        return None

    async def update(self, *, name: Optional[str] = None) -> Dict:
        """Update the dataset with specified fields.

        Args:
            name (str, optional): The new name for the dataset

        Returns:
            dict: The updated dataset
        """
        # Check by id
        existing_dataset_by_id = self._find_or_create_client_by_id_or_name(
            memory_storage_client=self._memory_storage_client,
            id=self._id,
            name=self._name,
        )

        if existing_dataset_by_id is None:
            _raise_on_non_existing_storage(_StorageTypes.DATASET, self._id)

        # Skip if no changes
        if name is None:
            return existing_dataset_by_id._to_resource_info()

        async with existing_dataset_by_id._file_operation_lock:
            # Check that name is not in use already
            existing_dataset_by_name = next(
                (dataset for dataset in self._memory_storage_client._datasets_handled if dataset._name and dataset._name.lower() == name.lower()),
                None,
            )

            if existing_dataset_by_name is not None:
                _raise_on_duplicate_storage(_StorageTypes.DATASET, 'name', name)

            existing_dataset_by_id._name = name

            previous_dir = existing_dataset_by_id._resource_directory

            existing_dataset_by_id._resource_directory = os.path.join(self._memory_storage_client._datasets_directory, name)

            await _force_rename(previous_dir, existing_dataset_by_id._resource_directory)

            # Update timestamps
            await existing_dataset_by_id._update_timestamps(True)

        return existing_dataset_by_id._to_resource_info()

    async def delete(self) -> None:
        """Delete the dataset."""
        dataset = next((dataset for dataset in self._memory_storage_client._datasets_handled if dataset._id == self._id), None)

        if dataset is not None:
            async with dataset._file_operation_lock:
                self._memory_storage_client._datasets_handled.remove(dataset)
                dataset._item_count = 0
                dataset._dataset_entries.clear()

                if os.path.exists(dataset._resource_directory):
                    await aioshutil.rmtree(dataset._resource_directory)

    async def list_items(
        self,
        *,
        offset: Optional[int] = 0,
        limit: Optional[int] = LIST_ITEMS_LIMIT,
        clean: Optional[bool] = None,  # noqa: U100
        desc: Optional[bool] = None,
        fields: Optional[List[str]] = None,  # noqa: U100
        omit: Optional[List[str]] = None,  # noqa: U100
        unwind: Optional[str] = None,  # noqa: U100
        skip_empty: Optional[bool] = None,  # noqa: U100
        skip_hidden: Optional[bool] = None,  # noqa: U100
        flatten: Optional[List[str]] = None,  # noqa: U100
        view: Optional[str] = None,  # noqa: U100
    ) -> ListPage:
        """List the items of the dataset.

        Args:
            offset (int, optional): Number of items that should be skipped at the start. The default value is 0
            limit (int, optional): Maximum number of items to return. By default there is no limit.
            desc (bool, optional): By default, results are returned in the same order as they were stored.
                To reverse the order, set this parameter to True.
            clean (bool, optional): If True, returns only non-empty items and skips hidden fields (i.e. fields starting with the # character).
                The clean parameter is just a shortcut for skip_hidden=True and skip_empty=True parameters.
                Note that since some objects might be skipped from the output, that the result might contain less items than the limit value.
            fields (list of str, optional): A list of fields which should be picked from the items,
                only these fields will remain in the resulting record objects.
                Note that the fields in the outputted items are sorted the same way as they are specified in the fields parameter.
                You can use this feature to effectively fix the output format.
            omit (list of str, optional): A list of fields which should be omitted from the items.
            unwind (str, optional): Name of a field which should be unwound.
                If the field is an array then every element of the array will become a separate record and merged with parent object.
                If the unwound field is an object then it is merged with the parent object.
                If the unwound field is missing or its value is neither an array nor an object and therefore cannot be merged with a parent object,
                then the item gets preserved as it is. Note that the unwound items ignore the desc parameter.
            skip_empty (bool, optional): If True, then empty items are skipped from the output.
                Note that if used, the results might contain less items than the limit value.
            skip_hidden (bool, optional): If True, then hidden fields are skipped from the output, i.e. fields starting with the # character.
            flatten (list of str, optional): A list of fields that should be flattened
            view (str, optional): Name of the dataset view to be used

        Returns:
            ListPage: A page of the list of dataset items according to the specified filters.
        """
        # Check by id
        existing_dataset_by_id = self._find_or_create_client_by_id_or_name(
            memory_storage_client=self._memory_storage_client,
            id=self._id,
            name=self._name,
        )

        if existing_dataset_by_id is None:
            _raise_on_non_existing_storage(_StorageTypes.DATASET, self._id)

        async with existing_dataset_by_id._file_operation_lock:
            start, end = existing_dataset_by_id._get_start_and_end_indexes(
                max(existing_dataset_by_id._item_count - (offset or 0) - (limit or LIST_ITEMS_LIMIT), 0) if desc else offset or 0,
                limit,
            )

            items = []

            for idx in range(start, end):
                entry_number = self._generate_local_entry_name(idx)
                items.append(existing_dataset_by_id._dataset_entries[entry_number])

            await existing_dataset_by_id._update_timestamps(False)

            if desc:
                items.reverse()

            return ListPage({
                'count': len(items),
                'desc': desc or False,
                'items': items,
                'limit': limit or LIST_ITEMS_LIMIT,
                'offset': offset or 0,
                'total': existing_dataset_by_id._item_count,
            })

    async def iterate_items(
        self,
        *,
        offset: int = 0,
        limit: Optional[int] = None,
        clean: Optional[bool] = None,  # noqa: U100
        desc: Optional[bool] = None,
        fields: Optional[List[str]] = None,  # noqa: U100
        omit: Optional[List[str]] = None,  # noqa: U100
        unwind: Optional[str] = None,  # noqa: U100
        skip_empty: Optional[bool] = None,  # noqa: U100
        skip_hidden: Optional[bool] = None,  # noqa: U100
    ) -> AsyncIterator[Dict]:
        """Iterate over the items in the dataset.

        Args:
            offset (int, optional): Number of items that should be skipped at the start. The default value is 0
            limit (int, optional): Maximum number of items to return. By default there is no limit.
            desc (bool, optional): By default, results are returned in the same order as they were stored.
                To reverse the order, set this parameter to True.
            clean (bool, optional): If True, returns only non-empty items and skips hidden fields (i.e. fields starting with the # character).
                The clean parameter is just a shortcut for skip_hidden=True and skip_empty=True parameters.
                Note that since some objects might be skipped from the output, that the result might contain less items than the limit value.
            fields (list of str, optional): A list of fields which should be picked from the items,
                only these fields will remain in the resulting record objects.
                Note that the fields in the outputted items are sorted the same way as they are specified in the fields parameter.
                You can use this feature to effectively fix the output format.
            omit (list of str, optional): A list of fields which should be omitted from the items.
            unwind (str, optional): Name of a field which should be unwound.
                If the field is an array then every element of the array will become a separate record and merged with parent object.
                If the unwound field is an object then it is merged with the parent object.
                If the unwound field is missing or its value is neither an array nor an object and therefore cannot be merged with a parent object,
                then the item gets preserved as it is. Note that the unwound items ignore the desc parameter.
            skip_empty (bool, optional): If True, then empty items are skipped from the output.
                Note that if used, the results might contain less items than the limit value.
            skip_hidden (bool, optional): If True, then hidden fields are skipped from the output, i.e. fields starting with the # character.

        Yields:
            dict: An item from the dataset
        """
        cache_size = 1000
        first_item = offset

        # If there is no limit, set last_item to None until we get the total from the first API response
        if limit is None:
            last_item = None
        else:
            last_item = offset + limit

        current_offset = first_item
        while last_item is None or current_offset < last_item:
            if last_item is None:
                current_limit = cache_size
            else:
                current_limit = min(cache_size, last_item - current_offset)

            current_items_page = await self.list_items(
                offset=current_offset,
                limit=current_limit,
                desc=desc,
            )

            current_offset += current_items_page.count
            if last_item is None or current_items_page.total < last_item:
                last_item = current_items_page.total

            for item in current_items_page.items:
                yield item

    async def get_items_as_bytes(self, *_args: Any, **_kwargs: Any) -> bytes:  # noqa: D102
        raise NotImplementedError('This method is not supported in local memory storage.')

    async def stream_items(self, *_args: Any, **_kwargs: Any) -> AsyncIterator:  # noqa: D102
        raise NotImplementedError('This method is not supported in local memory storage.')

    async def push_items(self, items: JSONSerializable) -> None:
        """Push items to the dataset.

        Args:
            items: The items which to push in the dataset. Either a stringified JSON, a dictionary, or a list of strings or dictionaries.
        """
        # Check by id
        existing_dataset_by_id = self._find_or_create_client_by_id_or_name(
            memory_storage_client=self._memory_storage_client, id=self._id, name=self._name)

        if existing_dataset_by_id is None:
            _raise_on_non_existing_storage(_StorageTypes.DATASET, self._id)

        normalized = self._normalize_items(items)

        added_ids: List[str] = []
        for entry in normalized:
            existing_dataset_by_id._item_count += 1
            idx = self._generate_local_entry_name(existing_dataset_by_id._item_count)

            existing_dataset_by_id._dataset_entries[idx] = entry
            added_ids.append(idx)

        data_entries: List[Tuple[str, Dict]] = []
        for id in added_ids:
            data_entries.append((id, existing_dataset_by_id._dataset_entries[id]))

        async with existing_dataset_by_id._file_operation_lock:
            await existing_dataset_by_id._update_timestamps(True)

            await _update_dataset_items(
                data=data_entries,
                entity_directory=existing_dataset_by_id._resource_directory,
                persist_storage=self._memory_storage_client._persist_storage,
            )

    def _to_resource_info(self) -> Dict:
        """Retrieve the dataset info."""
        return {
            'id': self._id,
            'name': self._name,
            'itemCount': self._item_count,
            'accessedAt': self._accessed_at,
            'createdAt': self._created_at,
            'modifiedAt': self._modified_at,
        }

    async def _update_timestamps(self, has_been_modified: bool) -> None:
        """Update the timestamps of the dataset."""
        self._accessed_at = datetime.now(timezone.utc)

        if has_been_modified:
            self._modified_at = datetime.now(timezone.utc)

        dataset_info = self._to_resource_info()
        await _update_metadata(
            data=dataset_info,
            entity_directory=self._resource_directory,
            write_metadata=self._memory_storage_client._write_metadata,
        )

    def _get_start_and_end_indexes(self, offset: int, limit: Optional[int] = None) -> Tuple[int, int]:
        actual_limit = limit or self._item_count
        start = offset + 1
        end = min(offset + actual_limit, self._item_count) + 1
        return (start, end)

    def _generate_local_entry_name(self, idx: int) -> str:
        return str(idx).zfill(LOCAL_ENTRY_NAME_DIGITS)

    def _normalize_items(self, items: JSONSerializable) -> List[Dict]:
        def normalize_item(item: Any) -> Optional[Dict]:
            if type(item) is str:
                item = json.loads(item)

            if type(item) is list:
                received = ',\n'.join(item)
                raise ValueError(f'Each dataset item can only be a single JSON object, not an array. Received: [{received}]')

            if type(item) is not dict and item is not None:
                raise ValueError(f'Each dataset item must be a JSON object. Received: {item}')

            return item

        if type(items) is str:
            items = json.loads(items)

        result = list(map(normalize_item, items)) if type(items) is list else [normalize_item(items)]
        # filter(None, ..) returns items that are True
        return list(filter(None, result))

    @classmethod
    def _get_storages_dir(cls, memory_storage_client: 'MemoryStorageClient') -> str:
        return memory_storage_client._datasets_directory

    @classmethod
    def _get_storage_client_cache(cls, memory_storage_client: 'MemoryStorageClient') -> List['DatasetClient']:
        return memory_storage_client._datasets_handled

    @classmethod
    def _create_from_directory(
        cls,
        storage_directory: str,
        memory_storage_client: 'MemoryStorageClient',
        id: Optional[str] = None,
        name: Optional[str] = None,
    ) -> 'DatasetClient':
        item_count = 0
        created_at = datetime.now(timezone.utc)
        accessed_at = datetime.now(timezone.utc)
        modified_at = datetime.now(timezone.utc)
        entries: Dict[str, Dict] = {}

        has_seen_metadata_file = False

        # Access the dataset folder
        for entry in os.scandir(storage_directory):
            if entry.is_file():
                if entry.name == '__metadata__.json':
                    has_seen_metadata_file = True

                    # We have found the dataset's metadata file, build out information based on it
                    with open(os.path.join(storage_directory, entry.name)) as f:
                        metadata = json.load(f)
                    id = metadata['id']
                    name = metadata['name']
                    item_count = metadata['itemCount']
                    created_at = datetime.fromisoformat(metadata['createdAt'])
                    accessed_at = datetime.fromisoformat(metadata['accessedAt'])
                    modified_at = datetime.fromisoformat(metadata['modifiedAt'])

                    continue

                with open(os.path.join(storage_directory, entry.name)) as f:
                    entry_content = json.load(f)
                entry_name = entry.name.split('.')[0]

                entries[entry_name] = entry_content

                if not has_seen_metadata_file:
                    item_count += 1

        new_client = DatasetClient(base_storage_directory=memory_storage_client._datasets_directory,
                                   memory_storage_client=memory_storage_client, id=id, name=name)

        # Overwrite properties
        new_client._accessed_at = accessed_at
        new_client._created_at = created_at
        new_client._modified_at = modified_at
        new_client._item_count = item_count

        for entry_id, content in entries.items():
            new_client._dataset_entries[entry_id] = content

        return new_client
